from typing import List
from kedro.io import AbstractDataset
from strawberry.types import Info
from .base import SignedUrlProvider
from ..config import load_config
from .. permissions import get_permissions
from .local_file_provider import LocalFileProvider
from ..utils import parse_s3_filepath
from ..models import DataSet, SignedUrl, SignedUrls, SignedUrlField
from ..exceptions import DataSetConfigError, DataSetError
import boto3
from botocore.exceptions import ClientError
from ..logs.logger import logger

CONFIG = load_config()
logger.debug("configuration loaded by {s}".format(s=__name__))

PERMISSIONS_CLASS = get_permissions(CONFIG.get("KEDRO_GRAPHQL_PERMISSIONS"))
logger.info("{s} using permissions class: {d}".format(s=__name__, d=PERMISSIONS_CLASS))


class S3Provider(SignedUrlProvider):
    """
    Implementation of SignedUrlProvider for AWS S3.

    """

    @staticmethod
    def presigned_url(filepath: str, expires_in_sec: int) -> SignedUrl | None:
        """
        Generate a signed URL for an S3 file.
        Args:
            filepath (str): The S3 file path in the format s3://bucket-name/key
            expires_in_sec (int): The number of seconds the signed URL should be valid for.

        Returns:
            Optional[str]: A signed URL for the S3 file.
        """
        logger.info(f"Signing S3 URL for filepath: {filepath}")
        bucket_name, key, filename = parse_s3_filepath(filepath)

        s3_client = boto3.client('s3')
        params = {
            'Bucket': bucket_name,
            'Key': f"{key}/{filename}" if key else filename
        }

        try:
            return SignedUrl(url=s3_client.generate_presigned_url(
                'get_object',
                Params=params,
                ExpiresIn=expires_in_sec
            ),
                file=filename)
        except ClientError as e:
            logger.error(f"Failed to generate signed URL: {e}")
            return None

    @staticmethod
    def presigned_post(filepath: str, expires_in_sec: int) -> SignedUrl | None:
        """
        Generate a presigned POST URL for uploading to S3.
        Args:
            filepath (str): The S3 file path in the format s3://bucket-name/key
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
        Returns:
            Optional[dict]: A dictionary with the URL and form fields for the presigned POST.
        """
        logger.info(f"Creating presigned POST for S3 filepath: {filepath}")
        bucket_name, key, filename = parse_s3_filepath(str(filepath))

        s3_client = boto3.client('s3')
        try:
            resp = s3_client.generate_presigned_post(bucket_name,
                                                     f"{key}/{filename}",
                                                     Fields=None,
                                                     Conditions=None,
                                                     ExpiresIn=expires_in_sec)
            return SignedUrl(url=resp["url"], file=filename, fields=[SignedUrlField(name=k, value=v) for k, v in resp["fields"].items()])
        except ClientError as e:
            logger.error(f"Failed to generate signed URL: {e}")
            return None

    @staticmethod
    def read(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Generate a signed URL S3 to download a file.

        Args:
            filepath (str): The S3 file path in the format s3://bucket-name/key
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
            partitions (list | None): Optional list of partitions in a PartitionedDataset.

        Returns:
            str | List[str]: download url with query parameters

            Example: https://your-bucket-name.s3.amazonaws.com/your-object-key?AWSAccessKeyId=your-access-key-id&Signature=your-signature&x-amz-security-token=your-security-token&Expires=expiration-time
        """
        c = dataset.parse_config()

        if c["type"] in ["partitions.PartitionedDataset"]:

            protocol, path = dataset.parse_path()
            if protocol == "file":
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={path}, protocol=file, using LocalFileProvider")

                return LocalFileProvider.read(info, dataset, expires_in_sec, partitions)

            elif protocol != "s3":

                raise DataSetConfigError(
                    "Invalid dataset configuration. Must have 'filepath' with 's3' protocol")
            else:
                if not partitions or len(partitions) == 0:
                    # if no partitions provided,
                    # need the list of files in the partitioned dataset to generate signed URLs for each partition
                    files = AbstractDataset.from_config(dataset.name, c).load()
                    files = [path + "/" + f + c.get("filename_suffix", "")
                             for f in files]
                else:
                    files = [path + "/" + partition + c.get("filename_suffix", "")
                             for partition in partitions]
                signed_urls = []
                for file in files:
                    logger.info(
                        f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={file}, protocol=s3, expires_in_sec={expires_in_sec}")

                    signed = S3Provider.presigned_url(
                        file, expires_in_sec)

                    signed_urls.append(signed)
                return SignedUrls(urls=signed_urls)

        elif c["type"] in ["partitions.IncrementalDataset"]:
            raise DataSetError(
                "Signed URLs for reading IncrementalDatasets are not supported.")
        else:

            protocol, filepath = dataset.parse_filepath()
            if protocol == "file":
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={filepath}, protocol=file, using LocalFileProvider")

                return LocalFileProvider.read(info, dataset, expires_in_sec, partitions)

            elif protocol != "s3":

                raise DataSetConfigError(
                    "Invalid dataset configuration. Must have 'filepath' with 's3' protocol")
            else:

                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={filepath}, protocol=s3, expires_in_sec={expires_in_sec}")
                return S3Provider.presigned_url(
                    filepath, expires_in_sec)

    @staticmethod
    def create(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Generate a signed URL S3 to upload a file.

        Args:
            info (Info): Strawberry GraphQL Info object.
            dataset (DataSet): The dataset to create a signed URL for.
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
            partitions (list | None): Optional list of partitions in a PartitionedDataset.

        Returns:
            dict | List[dict]: Dictionary with the URL to post to and form fields and values to submit with the POST. If an error occurs, returns None.

            Example:
                {
                    "url": "https://your-bucket-name.s3.amazonaws.com/",
                    "fields": {
                        "key": "your-object-key",
                        "AWSAccessKeyId": "your-access-key-id",
                        "x-amz-security-token": "your-security-token",
                        "policy": "your-policy",
                        "signature": "your-signature"
                    }
                }
        """
        c = dataset.parse_config()

        if c["type"] in ["partitions.PartitionedDataset"]:

            protocol, path = dataset.parse_path()
            if protocol == "file":
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath={path}, protocol=file, using LocalFileProvider")

                return LocalFileProvider.create(info, dataset, expires_in_sec, partitions)

            elif protocol != "s3":

                raise DataSetConfigError(
                    "Invalid dataset configuration. Must have 'filepath' with 's3' protocol")
            else:
                if not partitions or len(partitions) == 0:
                    # if no partitions provided,
                    # need the list of files in the partitioned dataset to generate signed URLs for each partition
                    files = AbstractDataset.from_config(dataset.name, c).load()
                    files = [path + "/" + f + c.get("filename_suffix", "")
                             for f in files]
                else:
                    files = [path + "/" + partition + c.get("filename_suffix", "")
                             for partition in partitions]
                signed_urls = []
                for file in files:
                    logger.info(
                        f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath={file}, protocol=s3, expires_in_sec={expires_in_sec}")

                    signed = S3Provider.presigned_post(
                        file, expires_in_sec)

                    signed_urls.append(signed)
                return SignedUrls(urls=signed_urls)

        elif c["type"] in ["partitions.IncrementalDataset"]:
            raise DataSetError(
                "Signed URLs for reading IncrementalDatasets are not supported.")
        else:
            protocol, filepath = dataset.parse_filepath()

            if protocol == "file":
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath={filepath}, protocol=file, using LocalFileProvider")

                return LocalFileProvider.create(info, dataset, expires_in_sec, partitions)

            elif protocol != "s3":

                raise DataSetConfigError(
                    "Invalid dataset configuration. Must have 'filepath' with 's3' protocol")
            else:
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath=s3://{filepath}, protocol=s3, expires_in_sec={expires_in_sec}")

                return S3Provider.presigned_post(filepath, expires_in_sec)



from importlib import import_module

from .logs.logger import logger

RESOLVER_PLUGINS = {}

TYPE_PLUGINS = {"query": [],
                "mutation": [],
                "subscription": []}


def discover_plugins(config):
    # discover plugins e.g. decorated functions e.g @gql_query, etc...
    for i in config["KEDRO_GRAPHQL_IMPORTS"]:
        import_module(i)


class NameConflictError(BaseException):
    """Raise for errors in adding plugins do to the same name."""


def gql_resolver(name):

    if name in RESOLVER_PLUGINS:
        raise NameConflictError(
            f"Plugin name conflict: '{name}'. Double check"
            " that all plugins have unique names."
        )

    def register_plugin(plugin_class):
        plugin = plugin_class()
        RESOLVER_PLUGINS[name] = plugin
        logger.info("registered resolver plugin '" + name + "' " + str(RESOLVER_PLUGINS[name]))
        return plugin

    return register_plugin


def gql_query():

    def register_plugin(plugin_class):
        TYPE_PLUGINS["query"].append(plugin_class)
        logger.info("registered type plugin 'query': " + str(plugin_class))
        return plugin_class

    return register_plugin


def gql_mutation():

    def register_plugin(plugin_class):
        TYPE_PLUGINS["mutation"].append(plugin_class)
        logger.info("registered type plugin 'query': " + str(plugin_class))
        return plugin_class

    return register_plugin


def gql_subscription():

    def register_plugin(plugin_class):
        TYPE_PLUGINS["subscription"].append(plugin_class)
        logger.info("registered type plugin 'query': " + str(plugin_class))
        return plugin_class

    return register_plugin



"""
This is a boilerplate pipeline 'example00'
generated using Kedro 0.18.4
"""

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import echo


def create_pipeline(**kwargs) -> Pipeline:
    return pipeline([
        node(
            func=echo,
            inputs=["text_in", "params:example", "parameters"],
            outputs="text_out",
            name="echo_node"
        )
    ])



import panel as pn
import param
import pandas as pd
from kedro_graphql.client import KedroGraphqlClient
from kedro_graphql.models import Pipeline

pn.extension('tabulator', css_files=[
             "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css"])


class PipelineMonitor(param.Parameterized):
    """Model for the PipelineMonitor component that handles the logic for fetching pipeline status and logs."""
    pipeline = param.ClassSelector(class_=Pipeline)
    client = param.ClassSelector(class_=KedroGraphqlClient)
    status = param.String()
    logs = param.String()

    def __init__(self, spec=None, pipeline=None, **params):
        """Initializes the Kedro GraphQL client with the provided specification.

        Kwargs:
            spec (dict): The specification for the UI, including configuration and pages.
            pipeline (Pipeline): The Kedro pipeline for which the monitor is built.
            **params: Additional parameters for the PipelineMonitor component.

        """
        super().__init__(**params)

        if pn.state.cookies:
            cookies = pn.state.cookies
        else:
            cookies = None

        self.client = KedroGraphqlClient(
            uri_graphql=spec["config"]["client_uri_graphql"],
            uri_ws=spec["config"]["client_uri_ws"],
            cookies=cookies
        )
        self.pipeline = pipeline

    @param.depends('pipeline', watch=True)
    async def subscribe_to_pipeline_events(self):
        """Fetches the current status of the pipeline."""

        # subscribe to pipeline events
        async for event in self.client.pipeline_events(id=self.pipeline.id):
            self.status = event.status
        await self.client.close_sessions()

    @param.depends('pipeline', watch=True)
    async def subscribe_to_logs(self):
        """Builds a terminal that displays the logs of the pipeline in real-time.
        This method connects to the pipeline logs and updates the terminal as new log messages are received.

        Yields:
        pn.widgets.Terminal: A terminal displaying the pipeline logs.
        """

        async for message in self.client.pipeline_logs(id=self.pipeline.id):
            if not self.logs:
                self.logs = ""
            self.logs += f"{message.time} - {message.message}\n"
        await self.client.close_sessions()

    def __panel__(self):
        return pn.Column(
            pn.Param(
                self,
                name="",
                parameters=["status", "logs"],
                widgets={
                    "status": pn.widgets.StaticText,
                    "logs": pn.widgets.TextAreaInput(
                        name="Live Log Stream",
                        sizing_mode='stretch_width',
                        height=400
                    )
                }
            ),
            sizing_mode='stretch_width'
        )



# some code commented out because it has not been implemented
import panel as pn
import param
import json


class PipelineViz(pn.viewable.Viewer):
    spec = param.Dict(default={})
    pipeline = param.String(default="")
    sid = param.String(default=None)
    height = param.Integer(default=800)

    def __init__(self, **params):
        super().__init__(**params)
        # self.viz_json = self.load_viz_json()

    def load_viz_json(self):
        """Not currently used, but can be used to load a static viz JSON file.
        This will be useful to programmtically hightlight nodes in the visualization,
        however, the kedro-graphql api does not currently support emitting events that 
        indicate which nodes are currently being processed.

        Returns:
            dict: The JSON data for the visualization.
        """
        with open(self.spec["config"]["viz_static"] + "api/main") as f:
            data = json.load(f)
        return data

    @param.depends("pipeline")
    async def build_viz(self):
        """Builds the visualization iframe for the specified pipeline.

        Returns:
            pn.pane.HTML: The HTML pane containing the iframe for the pipeline visualization.
        """
        iframe = """
        <iframe frameBorder="0" style="height:100%; width:100%" src="{prefix}pipeline/viz-build/index.html?pid={pipeline}"></iframe>
        """.format(pipeline=self.pipeline, prefix=self.spec["panel_get_server_kwargs"]["prefix"])
        if self.sid:
            iframe = """
            <iframe frameBorder="0" style="height:100%; width:100%" src="{prefix}pipeline/viz-build/index.html?pid={pipeline}&sid={sid}"></iframe>
            """.format(
                pipeline=self.pipeline,
                sid=self.sid,
                prefix=self.spec["panel_get_server_kwargs"]["prefix"]
            )

        yield pn.pane.HTML(iframe, height=self.height, sizing_mode="stretch_width")

    def __panel__(self):
        pn.state.location.sync(self, {"pipeline": "pipeline", "sid": "sid"})
        return pn.Card(
            self.build_viz,
            title="Pipeline Visualization",
            sizing_mode="stretch_width",
        )



# some code commented out because it has not been implemented
import panel as pn
import param
import pandas as pd
import json
from kedro_graphql.models import Pipelines

pn.extension('ace', 'jsoneditor')
pn.extension('tabulator', css_files=[
             "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css"])


class PipelineSearchModel(param.Parameterized):
    """Parameters for the PipelineSearch component."""
    spec = param.Dict(default={},
                      doc="The specification for the UI, including configuration and client.")
    search = param.String(
        default="", doc="Enter search terms to find relevant pipelines.")
    results_per_page = param.Integer(default=10, bounds=(
        1, 100), doc="Number of results per page.")
    # show_lineage = param.Boolean(
    #     default=False, doc="Whether to show lineage information.")
    load_more = param.Action(lambda x: x.load_more_results(),
                             label="Load More",
                             doc="Action to load more results.")
    load_prev = param.Action(lambda x: x.load_prev_results(),
                             label="Load Previous",
                             doc="Action to load previous results.")
    filter = param.String(
        default="", doc="Computed filter string for searching pipelines")

    cursor = param.String(default=None, doc="Cursor for pagination.")
    cursors = param.List(
        default=[], doc="List of cursors for pagination.")
    result = param.ClassSelector(
        class_=Pipelines, default=None, doc="The result of the pipeline search.")
    search_params = param.Dict(default={})

    @param.depends('search', watch=True)
    def compute_filter(self):
        """Compute the filter string based on the search input.

        Returns:
            str: A filter string in JSON format e.g. "{\"tags.key\": \"unique\", \"tags.value\": \"unique\"}"
        """
        filter = {"$or": []}
        parts = self.search.split(",")
        names = [p for p in parts if "name" in p]
        status = [p for p in parts if "state" in p]
        tags = [p for p in parts if "tag" in p]

        for name in names:
            if len(name.split("=")) > 1:
                n = name.split("=")[1]
                f = {"name": {"$regex": "^" + n}}
                filter["$or"].append(f)

        for s in status:
            if len(s.split("=")) > 1:
                n = s.split("=")[1]
                f = {"status.state": {"$regex": "^" + n}}
                filter["$or"].append(f)

        for tag in tags:
            if len(tag.split(":")) > 1:
                key_value = tag.split(":")[1]
                key = key_value.split("=")[0]
                value = key_value.split("=")[1]
                t = {
                    "tags.key": {"$regex": "^" + key},
                    "tags.value": {"$regex": "^" + value}
                }
                filter["$or"].append(t)

        if len(filter["$or"]) == 0 and len(self.search) > 0:
            # default to searching all fields for the raw string
            f = {
                "$or": [
                    {"name": {"$regex": "^" + self.search}},
                    {"status.state": {"$regex": "^" + self.search}},
                    {"tags.key": {"$regex": "^" + self.search}},
                    {"tags.value": {"$regex": "^" + self.search}}
                ]
            }
            self.filter = json.dumps(f)
        elif len(self.search) == 0:
            self.filter = ""
        else:
            self.filter = json.dumps(filter)

    @param.depends('search', watch=True)
    async def reset_cursor_history(self):
        """Reset the cursor history when the search input changes."""
        self.cursors = []
        self.cursor = ""

    @param.depends('load_more', watch=True)
    def load_more_results(self):
        """Load more results based on the current cursor."""
        if self.result.page_meta and self.result.page_meta.next_cursor:
            if self.cursor:
                if len(self.cursors) > 0 and self.cursors[-1] != self.cursor:
                    self.cursors.append(self.cursor)
                else:
                    self.cursors = [self.cursor]
                # print("CURSORS", self.cursors)
            # print("NEXT CURSOR", self.result.page_meta.next_cursor)
            self.cursor = self.result.page_meta.next_cursor

    @param.depends('load_prev', watch=True)
    def load_prev_results(self):
        """Load previous results based on the cursor history."""
        if len(self.cursors) > 0:
            self.cursor = self.cursors.pop()
            # print("CURSORS", self.cursors)
        else:
            self.cursor = ""

    @param.depends('results_per_page', 'cursor', 'filter', watch=True)
    async def build_search_params(self):
        """Build the search parameters for the API request."""
        if not self.cursor or len(self.cursor) == 0:
            cursor = None
        else:
            cursor = self.cursor
        self.search_params = {
            "limit": self.results_per_page,
            "cursor": cursor,
            "filter": self.filter
        }

    @param.depends('load_more_results', 'load_prev_results', 'build_search_params', watch=True)
    async def execute_search(self):
        """Execute the search based on the current filter and pagination parameters."""
        # print("Executing search with params:", self.search_params)
        self.result = await self.spec["config"]["client"].read_pipelines(
            limit=self.search_params.get("limit", self.results_per_page),
            cursor=self.search_params.get("cursor", self.cursor),
            filter=self.search_params.get("filter", self.filter)
        )


class PipelineSearch(pn.viewable.Viewer):
    """A component that allows users to search for pipelines in the Kedro GraphQL UI.
    This component provides a search bar, results per page selection, and buttons to load more or previous results.
    It displays the results in a table format with clickable rows to navigate to the pipeline details.

    Attributes:
        spec (dict): The specification for the UI, including configuration and client.
        limit (int): The number of results to display per page.
        cursor (str): The cursor for pagination.
        filter (str): The filter string for searching pipelines.
        prev_cursor (str): The cursor for the previous page.
        next_cursor (str): The cursor for the next page.
        more_clicks (int): The number of clicks on the "Load More" button.
        prev_clicks (int): The number of clicks on the "Load Previous" button.
        cursors (list): A list of cursors for pagination.
        cursor_index (int): The index of the current cursor in the cursors list.
        dashboard_page (str): The page to navigate to when a pipeline is clicked.
    """
    spec = param.Dict(default={})
    search_model = param.ClassSelector(
        class_=PipelineSearchModel, default=PipelineSearchModel())
    dashboard_page = param.String(
        default="dashboard", doc="The page to navigate to when a pipeline is clicked.")

    def __init__(self, **params):
        super().__init__(**params)
        self.search_model = PipelineSearchModel(spec=self.spec)

    def navigate(self, event, df):
        """Navigate to the pipeline details page when a row is clicked.
        Args:
            event (pn.widgets.Tabulator.ClickEvent): The event triggered by clicking a row.
            df (pd.DataFrame): The DataFrame containing pipeline data.
        """
        if event.column == "Open":
            pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + self.dashboard_page.lower()
            pn.state.location.search = "?pipeline=" + df.loc[event.row, "name"] + "&id=" + df.loc[event.row, "id"]
            pn.state.location.reload = True

    @param.depends('search_model.result')
    async def build_table(self):
        """Builds a table of pipelines based on the provided filter and pagination parameters.
        Args:
            limit (int): The number of results to display per page.
            filter (str): The filter string for searching pipelines.
            load_more (int): The number of clicks on the "Load More" button.
            load_prev (int): The number of clicks on the "Load Previous" button.
            show_lineage (list): A list indicating whether to show lineage in the table.
        Yields:
            pn.widgets.Tabulator: A table displaying the pipelines with clickable rows.
        """
        if not self.search_model.result or not self.search_model.result.pipelines:
            yield pn.pane.Markdown("No pipelines found matching the search criteria.")
        else:
            pipelines = self.search_model.result.pipelines

            df = pd.DataFrame.from_records([p.encode(encoder="dict") for p in pipelines])

            # flatten tags into seperate columns
            tags = []
            for index, row in df.iterrows():
                inner_tags = []
                if row["tags"]:
                    for i in row["tags"]:
                        inner_tags.append(
                            pd.DataFrame({"tag: " + i["key"]: i["value"]}, index=[index])
                        )

                    tags.append(pd.concat(inner_tags, axis=1))
                else:
                    tags.append(pd.DataFrame(index=[index]))

            tags = pd.concat(tags).fillna("")

            df = df.merge(tags, left_index=True, right_index=True)

            # parse latest state
            df["state"] = df["status"].apply(lambda x: x[-1]["state"])

            # visible columns
            cols_show = ["parent", "id", "name", "state"] + list(tags.columns)

            df_widget = pn.widgets.Tabulator(
                df[cols_show],
                buttons={'Open': "<i class='fa fa-folder-open'></i>"},
                theme='materialize',
                hidden_columns=["parent"],
                disabled=True,
                sizing_mode="stretch_width",
                show_index=False
            )
            df_widget.on_click(lambda e: self.navigate(e, df))
            yield df_widget

    def __panel__(self):
        pn.state.onload(self.search_model.execute_search)
        return pn.Card(
            pn.Row(
                pn.pane.Markdown("""Search for pipelines with the following syntax:
- Search by name: `name=example`
- Search by status: `state=STAGED`
- Search by tags: `tag:key=value`
- Search by multiple tags: `tag:key1=value1,tag:key2=value2`
- Search by multiple fields: `name=example,state=STAGED,tag:key=value`
- Search by pipeline id: `id=67ca04f127d86d04151c90eb`
                                 """)
            ),
            pn.Row(
                pn.Param(
                    self.search_model,
                    name="",
                    parameters=["search", "results_per_page", "load_more", "load_prev"],
                    show_name=False,
                    display_threshold=0,
                    default_layout=pn.Row,
                    widgets={
                        "load_more_clicks": pn.widgets.Button,
                        "load_prev_clicks": pn.widgets.Button,
                    }
                )
            ),
            pn.Row(
                self.build_table,
                sizing_mode="stretch_width"
            ),
            sizing_mode="stretch_width"
        )



import asyncio
import param
import panel as pn
from kedro_graphql.client import KedroGraphqlClient
from kedro_graphql.models import Pipeline, PipelineInput, ParameterInput, DataSetInput, TagInput, PipelineInputStatus, PipelineSlice

pn.extension('modal')
pn.extension(notifications=True)


class PipelineRetry(pn.viewable.Viewer):
    """A Panel component that retries execution of a Kedro pipeline or a slice of it.

    Attributes:
        pipeline (Pipeline): The Kedro pipeline to retry.
        client (KedroGraphqlClient): The client used to interact with the Kedro GraphQL API.
    """

    pipeline = param.ClassSelector(class_=Pipeline)
    client = param.ClassSelector(class_=KedroGraphqlClient)
    select_value = param.Selector(default='Whole', objects=['Whole', 'Slice'], doc="Select retry strategy")

    def __init__(self, **params):
        super().__init__(**params)

        pn.state.notifications.position = 'top-center'

        if not self.pipeline or not self.client:
            self._content = pn.Column(
                pn.pane.Markdown("**Missing required parameters: `pipeline`, or `client`.**"),
                sizing_mode="stretch_width"
            )
        else:
            # Show loading spinner until the dashboard is built
            self._content = pn.Column(
                pn.indicators.LoadingSpinner(value=True, width=50, height=50),
                pn.pane.Markdown("Fetching Data..."),
                sizing_mode='stretch_width'
            )
            
            # Ensures build after panel server is fully loaded to avoid race conditions
            pn.state.onload(lambda: asyncio.create_task(asyncio.to_thread(self.build_component)))
    
    def _on_select_change(self, event):
        """Handle select widget value changes by updating the parameter."""
        self.select_value = event.new

    @param.depends('select_value', watch=True)
    def build_component(self):
        """
        Builds a card for retrying the pipeline or a slice of it based on the selected strategy.
        """
        # Create the select widget and sync it with the parameter
        self.select = pn.widgets.Select(
            name='Select retry strategy',
            options=['Whole', 'Slice'],
            value=self.select_value,
            sizing_mode="stretch_width"
        )
        self.select.param.watch(self._on_select_change, 'value')
        
        self.only_missing = pn.widgets.Checkbox(name='Only Missing')

        self.slice_inputs = {
            "tags": pn.widgets.TextInput(
                placeholder=f"e.g. tag1,tag2",
                name="Tags",
                sizing_mode="stretch_width"
            ),
            "from_nodes": pn.widgets.TextInput(
                placeholder="e.g. node_name1,node_name2",
                name="From Nodes",
                sizing_mode="stretch_width"
            ),
            "to_nodes": pn.widgets.TextInput(
                placeholder="e.g. node_name3,node_name4",
                name="To Nodes",
                sizing_mode="stretch_width"
            ),
            "node_names": pn.widgets.TextInput(
                placeholder="e.g. node_name5,node_name6",
                name="Node Names",
                sizing_mode="stretch_width"
            ),
            "from_inputs": pn.widgets.TextInput(
                placeholder="e.g. input1,input2",
                name="From Inputs",
                sizing_mode="stretch_width"
            ),
            "to_outputs": pn.widgets.TextInput(
                placeholder="e.g. output1,output2",
                name="To Outputs",
                sizing_mode="stretch_width"
            ),
            "node_namespace": pn.widgets.TextInput(
                placeholder="e.g. namespace1,namespace2",
                name="Node Namespace",
                sizing_mode="stretch_width"
            ),
        }

        yes_button = pn.widgets.Button(name="Yes", button_type="primary")
        cancel_button = pn.widgets.Button(name="Cancel")
        yes_button.on_click(lambda e: asyncio.create_task(self.pipeline_run()))
        cancel_button.on_click(lambda e: self.modal.hide())

        self.modal = pn.Modal(
            pn.Column(
                pn.pane.Markdown("### Are you sure you want to retry? This could overwrite any previous data."),
                pn.Row(yes_button, cancel_button)
            ),
            name="Confirm Retry",
            margin=20,
            width=400,
        )

        if self.select_value == 'Whole':
            retry_all_button = pn.widgets.Button(
                name="Retry Whole Pipeline",
                button_type="success",
                sizing_mode="stretch_width",
            )
            retry_all_button.on_click(lambda event: self.modal.toggle())
            card = pn.Card(
                self.select,
                retry_all_button,
                title="Retry Whole Pipeline",
                width=400,
            )
        else:
            retry_slice_button = pn.widgets.Button(
                name="Retry Pipeline Slice",
                button_type="success",
                sizing_mode="stretch_width",
            )
            retry_slice_button.on_click(lambda event: self.modal.toggle())
            card = pn.Card(
                self.select,
                pn.Row(
                    self.slice_inputs["tags"],
                    pn.widgets.TooltipIcon(value="Construct the pipeline using only nodes which have this tag attached.")
                ),
                pn.Row(
                    self.slice_inputs["from_nodes"],
                    pn.widgets.TooltipIcon(value="A list of node names which should be used as a starting point.")
                ),
                pn.Row(
                    self.slice_inputs["to_nodes"],
                    pn.widgets.TooltipIcon(value="A list of node names which should be used as an end point.")
                ),
                pn.Row(
                    self.slice_inputs["node_names"],
                    pn.widgets.TooltipIcon(value="Run only nodes with specified names.")
                ),
                pn.Row(
                    self.slice_inputs["from_inputs"],
                    pn.widgets.TooltipIcon(value="A list of dataset names which should be used as a starting point.")
                ),
                pn.Row(
                    self.slice_inputs["to_outputs"],
                    pn.widgets.TooltipIcon(value="A list of dataset names which should be used as an end point.")
                ),
                pn.Row(
                    self.slice_inputs["node_namespace"],
                    pn.widgets.TooltipIcon(value="Name of the node namespace to run.")
                ),
                pn.Row(
                    self.only_missing,
                    pn.widgets.TooltipIcon(value="Run only the missing outputs.")
                ),
                retry_slice_button,
                title="Retry a Slice",
                width=400
            )
        
        self._content[:] = pn.Column(
            card,
            self.modal,
        )

    async def pipeline_run(self):
        """Run the pipeline based on the selected retry strategy."""

        self.modal.hide()

        try:
            if self.select_value == 'Whole':
                pipeline_input = PipelineInput(
                    name=self.pipeline.name,
                    parameters=[
                        ParameterInput(
                            name=param.name,
                            value=param.value,
                            type=param.type.name
                        ) for param in self.pipeline.parameters
                    ] if self.pipeline.parameters else [],
                    data_catalog=[
                        DataSetInput(
                            name=ds.name,
                            config=ds.config,
                            tags=[
                                TagInput(key=tag.key, value=tag.value) for tag in ds.tags
                            ] if ds.tags else []
                        ) for ds in self.pipeline.data_catalog
                    ] if self.pipeline.data_catalog else [],
                    tags=[
                        TagInput(key=tag.key, value=tag.value) for tag in self.pipeline.tags
                    ] if self.pipeline.tags else [],
                    runner=self.pipeline.status[-1].runner,
                    state=PipelineInputStatus.READY
                )
            else:
                slices = []
                for key, field in self.slice_inputs.items():
                    if field.value.strip():
                        slices.append(
                            PipelineSlice(
                                slice=key.upper(),
                                args=[item.strip() for item in field.value.split(',')]
                            )
                        )

                pipeline_input = PipelineInput(
                    name=self.pipeline.name,
                    parameters=[
                        ParameterInput(
                            name=param.name,
                            value=param.value,
                            type=param.type.name
                        ) for param in self.pipeline.parameters
                    ] if self.pipeline.parameters else [],
                    data_catalog=[
                        DataSetInput(name=ds.name, config=ds.config)
                        for ds in self.pipeline.data_catalog
                    ] if self.pipeline.data_catalog else [],
                    tags=[
                        TagInput(key=tag.key, value=tag.value) for tag in self.pipeline.tags
                    ] if self.pipeline.tags else [],
                    runner=self.pipeline.status[-1].runner,
                    state=PipelineInputStatus.READY,
                    slices=slices,
                    only_missing=self.only_missing.value
                )

            await self.client.update_pipeline(id=self.pipeline.id, pipeline_input=pipeline_input)

        except Exception as e:
            pn.state.notifications.error(
                f"Failed to trigger {self.select_value.lower()} retry for {self.pipeline.name}: {str(e)}",
                duration=10000
            )
            return

        pn.state.notifications.success(
            f'Successfully triggered {self.select_value.lower()} retry for {self.pipeline.name}',
            duration=10000
        )

    def __panel__(self):
        return self._content



import gql
import json
from functools import reduce
from graphql.language import print_ast
from importlib_resources import files
from typing import Any, Optional
from .models import Pipeline
from urllib.parse import urlparse
from .logs.logger import logger
from .exceptions import DataSetConfigError


def build_graphql_query(
    query_name: str, fragments: Optional[list] = None, query_file="queries.gql"
) -> str:

    query_path = files("kedro_graphql.static").joinpath(query_file)

    # ensure it exists
    if not query_path.exists():
        raise FileNotFoundError(
            f"Query file {query_file} not found. Ensure it exists in kedro_graphql.static"
        )

    with open(query_path, "r") as f:
        document = gql.gql(f.read())  # Parse the entire document

    outstr = ""

    # # Iterate through definitions and extract named operations
    for definition in document.definitions:
        if hasattr(definition, "name") and definition.name.value == query_name:
            outstr += print_ast(definition) + "\n"

    if not outstr:
        raise ValueError(f"Query {query_name} not found in {query_file}")

    if fragments:
        count = 0
        for definition in document.definitions:
            if hasattr(definition, "name") and definition.name.value in fragments:
                outstr += print_ast(definition) + "\n"
                count += 1
        if count != len(fragments):
            raise ValueError(f"Not all fragments {fragments} found in {query_file}")

    return outstr


def merge(a, b, path=None):
    """
    Merges nested dictionaries recursively.  Merges b into a.
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def merge_dicts(dicts):
    return reduce(merge, dicts)


def parse_s3_filepath(filepath: str) -> tuple[str, str]:
    """
    Parse the s3 bucket name and key from DataSet filepath field.

    Args:
        filepath (str): The S3 file path in the format s3://bucket-name/key

    Returns:
        tuple: A tuple containing the bucket name, the S3 prefix, and the S3 object name.

    Raises:
        ValueError: If the filepath does not start with "s3://" or if the bucket name or S3 key is missing.
    """

    if not filepath.startswith("s3://"):
        raise ValueError("Invalid S3 path. Must start with 's3://'")

    parsed = urlparse(filepath)
    bucket_name = parsed.netloc
    s3_key = parsed.path.lstrip("/")
    filename = s3_key.split("/")[-1]
    s3_key = "/".join(s3_key.split("/")[:-1])  # Remove the filename from the key

    if not bucket_name:
        raise ValueError("Invalid S3 path. Bucket name is missing.")
    # if not s3_key:
    #    raise ValueError("Invalid S3 path. S3 key (object path) is missing.")

    return bucket_name, s3_key, filename


def add_param_to_feed_dict(feed_dict, param_name: str, param_value: Any, add_prefix=True) -> None:
    """Context-free version of utility found inside KedroContext._get_feed_dict method. Option to add params: prefix if desired.

    Example:

        >>> param_name = "a"
        >>> param_value = {"b": 1}
        >>> feed_dict = {}
        >>> _add_param_to_feed_dict(feed_dict, param_name, param_value)
        >>> assert feed_dict["params:a"] == {"b": 1}
        >>> assert feed_dict["params:a.b"] == 1
    """
    key = f"params:{param_name}" if add_prefix else param_name
    feed_dict[key] = param_value
    if isinstance(param_value, dict):
        for key, val in param_value.items():
            add_param_to_feed_dict(feed_dict, f"{param_name}.{key}", val)


def generate_unique_paths(pipeline: Pipeline, datasets: list) -> Pipeline:
    """Modifies filepaths in the pipeline's data catalog to ensure they are unique.

    Args:
        pipeline (Pipeline): The Kedro pipeline to scope filepaths for.
        datasets (list): List of dataset names to create unique filepaths for.

    Returns:
        Pipeline: The same pipeline with a modified data catalog containing unique filepaths for specified datasets.
    Raises:
        DataSetConfigError: If a dataset does not have a 'filepath' or 'path' key in its configuration.
    """
    for d in pipeline.data_catalog:
        if d.name in datasets:
            c = json.loads(d.config)
            if c.get("filepath", None):
                filepath = c["filepath"]
                parts = filepath.rsplit("/", 1)
                new_path = parts[0] + "/" + pipeline.id + "/" + parts[1]
                c["filepath"] = new_path
                logger.info(
                    f"Modifying dataset {d.name} filepath to {new_path} to ensure uniqueness.")
                d.config = json.dumps(c)
            elif c.get("path", None):
                path = c["path"]
                parts = path.rsplit("/", 1)
                new_path = parts[0] + "/" + pipeline.id + "/" + parts[1]
                c["path"] = new_path
                logger.info(
                    f"Modifying dataset {d.name} path to {new_path} to ensure uniqueness.")
                d.config = json.dumps(c)
            else:
                raise DataSetConfigError(
                    f"Dataset {d.name} does not have a 'filepath' or 'path' key in its configuration.")
        else:
            logger.info(
                f"Dataset {d.name} not in specified datasets list, skipping filepath modification.")
    return pipeline



from panel.auth import GenericLoginHandler, log, _deserialize_state, decode_response_body, decode_token, OAuthLoginHandler
from panel.config import config
from panel.io.state import state
from tornado.httpclient import HTTPError as HTTPClientError, HTTPRequest, HTTPError
import urllib.parse as urlparse


class PKCELoginHandler(GenericLoginHandler):
    """
    Handles the PKCE login flow for Panel.
    """

    async def _fetch_access_token(
        self, client_id, redirect_uri=None, client_secret=None, code=None,
        refresh_token=None, username=None, password=None
    ):
        """
        Overrides the panel.auth.OAuthLoginHandler._fetch_access_token method to implement the PKCE flow.
        See https://github.com/holoviz/panel/issues/7979.

        Fetches the access token.

        Parameters
        ----------
        client_id:
          The client ID
        redirect_uri:
          The redirect URI
        code:
          The response code from the server
        client_secret:
          The client secret
        refresh_token:
          A token used for refreshing the access_token
        username:
          A username
        password:
          A password
        """
        log.debug("%s making access token request.", type(self).__name__)
        params = {
            'client_id': client_id,
            **self._EXTRA_TOKEN_PARAMS
        }
        if redirect_uri:
            params['redirect_uri'] = redirect_uri
        if self._SCOPE:
            params['scope'] = ' '.join(self._SCOPE)
        if code:
            params['code'] = code
            params['code_verifier'] = self.get_code_cookie()
        if refresh_token:
            refreshing = True
            params['refresh_token'] = refresh_token
            params['grant_type'] = 'refresh_token'
        else:
            refreshing = False
        if client_secret:
            params['client_secret'] = client_secret
        elif username:
            params.update(username=username, password=password)
        else:
            params['code_verifier'] = self.get_code_cookie()

        http = self.get_auth_http_client()

        # Request the access token.
        req = HTTPRequest(
            self._OAUTH_ACCESS_TOKEN_URL,
            method='POST',
            body=urlparse.urlencode(params),
            headers=self._API_BASE_HEADERS
        )
        try:
            response = await http.fetch(req)
        except HTTPClientError as e:
            log.debug("%s access token request failed.", type(self).__name__)
            self._raise_error(e.response, status=401)

        if not response.body or not (body := decode_response_body(response)):
            log.debug("%s token endpoint did not return a valid access token.",
                      type(self).__name__)
            self._raise_error(response)

        if 'access_token' not in body:
            if refresh_token:
                log.debug("%s token endpoint did not reissue an access token.",
                          type(self).__name__)
                return None, None, None
            self._raise_error(response, body, status=401)

        expires_in = body.get('expires_in')
        if expires_in:
            expires_in = int(expires_in)

        access_token, refresh_token = body['access_token'], body.get('refresh_token')
        if refreshing:
            # When refreshing the tokens we do not need to re-fetch the id_token or user info
            return None, access_token, refresh_token, expires_in
        elif id_token := body.get('id_token'):
            try:
                user = OAuthLoginHandler.set_auth_cookies(
                    self, id_token, access_token, refresh_token, expires_in)
            except HTTPError:
                pass
            else:
                log.debug("%s successfully obtained access_token and id_token.",
                          type(self).__name__)
                return user, access_token, refresh_token, expires_in

        user_headers = dict(self._API_BASE_HEADERS)
        if self._OAUTH_USER_URL:
            if self._access_token_header:
                user_url = self._OAUTH_USER_URL
                user_headers['Authorization'] = self._access_token_header.format(
                    body['access_token']
                )
            else:
                user_url = '{}{}'.format(self._OAUTH_USER_URL, body['access_token'])

            log.debug("%s requesting OpenID userinfo.", type(self).__name__)
            try:
                user_response = await http.fetch(user_url, headers=user_headers)
                id_token = decode_response_body(user_response)
            except HTTPClientError:
                id_token = None

        if not id_token:
            log.debug(
                "%s could not fetch user information, the token endpoint did not "
                "return an id_token and no OpenID user info endpoint was provided. "
                "Attempting to code access_token to resolve user information.",
                type(self).__name__
            )
            try:
                id_token = decode_token(body['access_token'])
            except Exception:
                log.debug("%s could not decode access_token.", type(self).__name__)
                self._raise_error(response, body, status=401)

        log.debug("%s successfully obtained access_token and userinfo.",
                  type(self).__name__)
        user = OAuthLoginHandler.set_auth_cookies(
            self, id_token, access_token, refresh_token, expires_in)
        return user, access_token, refresh_token, expires_in

    async def get(self):
        """
        Handles the GET request for the PKCE login flow.

        Copied and modified from panel.auth.CodeChallengeLoginHandler
        See https://github.com/holoviz/panel/issues/7979
        """
        code = self.get_argument("code", "")
        url_state = self.get_argument("state", "")
        redirect_uri = self._redirect_uri
        if not code or not url_state:
            self._authorize_redirect(redirect_uri)
            return

        cookie_state = self.get_state_cookie()
        if cookie_state != url_state:
            log.warning("OAuth state mismatch: %s != %s", cookie_state, url_state)
            raise HTTPError(400, "OAuth state mismatch")

        decoded_state = _deserialize_state(url_state)
        user = await self.get_authenticated_user(redirect_uri, config.oauth_key, url_state, client_secret=config.oauth_secret, code=code)
        if user is None:
            raise HTTPError(403)
        log.debug("%s authorized user, redirecting to app.", type(self).__name__)
        self.redirect(decoded_state.get('next_url', state.base_url))

    def _authorize_redirect(self, redirect_uri):
        state = self.get_state()
        self.set_state_cookie(state)
        code_verifier, code_challenge = self.get_code()
        self.set_code_cookie(code_verifier)
        params = {
            "client_id": config.oauth_key,
            "response_type": "code",
            "scope": ' '.join(self._SCOPE),
            "state": state,
            "response_mode": "query",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "redirect_uri": redirect_uri
        }
        query_params = urlparse.urlencode(params)
        self.redirect(f"{self._OAUTH_AUTHORIZE_URL}?{query_params}")



import panel as pn
import param
from kedro_graphql.ui.decorators import UI_PLUGINS


class PipelineFormFactory(pn.viewable.Viewer):
    """A factory for building forms for Kedro pipelines using registered @ui_form plugins.
    This component allows users to select a form for a specific pipeline and build the form dynamically.

    Attributes:
        form (str): The name of the form to build.
        pipeline (str): The name of the pipeline for which the form is built.
        options (list): A list of available forms for the selected pipeline.
        spec (dict): The specification for the UI, including configuration and pages."""

    form = param.String(default=None)
    pipeline = param.String(default=None)
    options = param.List(default=[])
    spec = param.Dict(default={})

    def __init__(self, **params):
        super().__init__(**params)

    @param.depends("pipeline")
    async def build_options(self):
        """Builds the options for the form selection based on the registered @ui_form plugins for the selected pipeline."""

        yield pn.indicators.LoadingSpinner(value=True, width=25, height=25)

        self.options = []
        if UI_PLUGINS["FORMS"].get(self.pipeline, None):
            for f in UI_PLUGINS["FORMS"][self.pipeline]:
                self.options = [f.__name__ for f in UI_PLUGINS["FORMS"][self.pipeline]]

        if not self.form or self.form not in self.options:
            self.form = self.options[0] if self.options else None

        yield pn.widgets.Select.from_param(self.param.form,
                                           name='Select a form', options=self.options, value=self.param.form)

    @param.depends("form", "pipeline", "spec")
    async def build_form(self):
        """Builds the form based on the selected form and pipeline, using the registered @ui_form plugins."""

        yield pn.indicators.LoadingSpinner(value=True, width=25, height=25)

        form = None
        for f in UI_PLUGINS["FORMS"][self.pipeline]:
            if f.__name__ == self.form:
                form = f
        f = pn.bind(form, spec=self.param.spec)
        yield f

    def __panel__(self):
        pn.state.location.sync(
            self, {"pipeline": "pipeline", "form": "form"})

        return pn.Column(
            pn.Row(self.build_options),
            pn.Row(self.build_form)
        )



import json
import pytest
from kedro_graphql.client import PIPELINE_GQL
from kedro_graphql.models import Pipeline
from pathlib import Path


class TestSchemaExtensions:

    @pytest.mark.asyncio
    async def test_pipeline_extension(self,
                                      mocker,
                                      mock_app,
                                      mock_info_context,
                                      mock_pipeline):
        """
        Test that the kedro_graphql.schema.PipelineExtension correctly masks filepaths in the pipeline's data catalog.
        """
        # parse filepath prefix from mock_pipeline datasets
        c = json.loads(mock_pipeline.data_catalog[0].config)
        prefix = "/" + c["filepath"].split("/", 2)[1] + "/"

        mocker.patch.dict("kedro_graphql.schema.CONFIG", {"KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS": [
            {"prefix": prefix, "mask": "/REDACTED/"}]})

        query = """
                    query readPipeline($id: String!) {
                      readPipeline(id: $id) """ + PIPELINE_GQL + """
                    }
                """

        resp = await mock_app.schema.execute(query, variable_values={"id": str(mock_pipeline.id)})
        assert resp.errors is None
        p = Pipeline.decode(resp.data["readPipeline"], decoder="graphql")
        for d in p.data_catalog:
            c = json.loads(d.config)
            if c.get("filepath"):
                assert not c["filepath"].startswith(prefix)
                assert c["filepath"].startswith("/REDACTED/")

    @pytest.mark.asyncio
    async def test_pipeline_input_extension_00(self,
                                               mocker,
                                               mock_app,
                                               mock_info_context,
                                               mock_text_in,
                                               mock_text_out):
        """
        Test that the kedro_graphql.schema.PipelineInputExtension correctly validates filepaths in the
        pipeline's data catalog against allowed roots.
        """

        mocker.patch.dict("kedro_graphql.schema.CONFIG", {
                          "KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS": ["./data/"], })

        query = """
            mutation createPipeline($pipeline: PipelineInput!, $uniquePaths: [String!]) {
              createPipeline(pipeline: $pipeline, uniquePaths: $uniquePaths) """ + PIPELINE_GQL + """
            }
        """
        resp = await mock_app.schema.execute(query,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                 {"name": "text_out", "config": json.dumps(
                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             },
                                                 "uniquePaths": ["text_in", "text_out"]})

        assert resp.errors is not None
        assert "not allowed" in resp.errors[0].message

    @pytest.mark.asyncio
    async def test_pipeline_input_extension_01(self,
                                               mocker,
                                               mock_app,
                                               mock_info_context,
                                               mock_text_in,
                                               mock_text_out):
        """
        Test that the kedro_graphql.schema.PipelineInputExtension correctly unmasks the input Dataset filepaths in the
        pipeline's data catalog and validates against allowed roots.  Check to make sure pipeline is created in the 
        backend with unmasked filepaths.
        """
        prefix = "/" + str(mock_text_in).split("/", 2)[1] + "/"
        mocker.patch.dict("kedro_graphql.schema.CONFIG", {
                          "KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS": [prefix], })

        mocker.patch.dict("kedro_graphql.schema.CONFIG", {"KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS": [
            {"prefix": prefix, "mask": "/REDACTED/"}]})

        query = """
            mutation createPipeline($pipeline: PipelineInput!, $uniquePaths: [String!]) {
              createPipeline(pipeline: $pipeline, uniquePaths: $uniquePaths) """ + PIPELINE_GQL + """
            }
        """
        resp = await mock_app.schema.execute(query,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in).replace("/tmp/", "/REDACTED/")})},
                                                                 {"name": "text_out", "config": json.dumps(
                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out).replace("/tmp/", "/REDACTED/")})}
                                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             },
                                                 "uniquePaths": ["text_in", "text_out"]})

        assert resp.errors is None

        # check response to make sure filepaths are masked
        p = Pipeline.decode(resp.data["createPipeline"], decoder="graphql")
        for d in p.data_catalog:
            c = json.loads(d.config)
            if c.get("filepath"):
                assert c["filepath"].startswith("/REDACTED/")
                assert not c["filepath"].startswith(prefix)

        # check backend to make sure filepaths are unmasked
        pipeline = mock_app.backend.read(id=p.id)
        for d in pipeline.data_catalog:
            c = json.loads(d.config)
            if c.get("filepath"):
                assert c["filepath"].startswith(prefix)
                assert not c["filepath"].startswith("/REDACTED/")



import pytest
import json


def test_parse_s3_filepath():
    from kedro_graphql.utils import parse_s3_filepath

    # Test with a valid S3 path
    s3_path = "s3://my-bucket/my-folder/my-file.csv"
    bucket, key, filename = parse_s3_filepath(s3_path)
    assert bucket == "my-bucket"
    assert key == "my-folder"
    assert filename == "my-file.csv"

    # Test with a path that has no folder structure
    s3_path = "s3://my-bucket/my-file.csv"
    bucket, key, filename = parse_s3_filepath(s3_path)
    assert bucket == "my-bucket"
    assert key == ""
    assert filename == "my-file.csv"

    # Test with an invalid S3 path
    with pytest.raises(ValueError):
        parse_s3_filepath("invalid-s3-path")


def test_unique_paths(mock_pipeline_staged):
    from kedro_graphql.utils import generate_unique_paths
    p = generate_unique_paths(mock_pipeline_staged, ["text_in", "text_out"])
    datasets = {d.name: json.loads(d.config) for d in p.data_catalog}

    assert datasets["text_in"]["filepath"] != "./data/01_raw/text_in.csv"
    assert datasets["text_in"]["filepath"] == "./data/01_raw/{id}/text_in.csv".format(
        id=p.id)
    assert datasets["text_out"]["filepath"] != "./data/01_raw/text_out.csv"
    assert datasets["text_out"]["filepath"] == "./data/01_raw/{id}/text_out.csv".format(
        id=p.id)


def test_unique_paths_partitioned_datasests():
    from kedro_graphql.utils import generate_unique_paths
    from kedro_graphql.models import Pipeline, DataSet

    mock_pipeline = Pipeline(id="00000001", name="test_pipeline", data_catalog=[DataSet(
        name="partitioned_dataset",
        config=json.dumps({
            "type": "partitions.PartitionedDataset",
            "path": "./data/02_intermediate/partitioned_dataset",
            "filename_suffix": ".txt",
            "dataset": {
                "type": "text.TextDataset"
            }
        })
    )])

    p = generate_unique_paths(
        mock_pipeline, ["partitioned_dataset"])

    datasets = {d.name: json.loads(d.config) for d in p.data_catalog}

    assert datasets["partitioned_dataset"]["path"] != "./data/02_intermediate/partitioned_dataset"
    assert datasets["partitioned_dataset"]["path"] == "./data/02_intermediate/{id}/partitioned_dataset".format(
        id=p.id)



import pytest


class TestSchemaQuery:

    @pytest.mark.asyncio
    async def test_pipeline(self, mock_app, mock_info_context, mock_pipeline):

        query = """
        query TestQuery($id: String!) {
          readPipeline(id: $id){
            id
          }
        }
        """
        resp = await mock_app.schema.execute(query, variable_values={"id": str(mock_pipeline.id)})
        assert resp.errors is None

    @pytest.mark.asyncio
    async def test_pipelines(self, mock_app, mock_info_context, mock_pipeline):

        query = """
        query TestQuery($limit: Int!, $filter: String!) {
          readPipelines(limit: $limit, filter: $filter) {
            pageMeta {
              nextCursor
            }
            pipelines {
              id
            }
          }
        }
        """
        resp = await mock_app.schema.execute(query, variable_values={"limit": 3, "filter": "{\"tags\": {\"key\": \"author\", \"value\": \"opensean\"}}"})
        assert resp.errors is None

    @pytest.mark.asyncio
    async def test_pipeline_templates(self, mock_app, mock_info_context):

        query = """
        query TestQuery($limit: Int!) {
          pipelineTemplates(limit: $limit) {
            pageMeta {
              nextCursor
            }
            pipelineTemplates {
              name
              describe
              inputs {
                name
              }
              nodes {
                name
                inputs
                outputs
                tags
              }
              outputs {
                name
              }
              parameters {
                name
                value
              }
            }
          }
        }
        """
        resp = await mock_app.schema.execute(query, variable_values={"limit": 5})

        assert resp.errors is None

    @pytest.mark.asyncio
    async def test_read_datasets(self, mock_app, mock_info_context, mock_pipeline):

        query = """
        query TestQuery($id: String!, $datasets: [DataSetInput!]!, $expires_in_sec: Int!) {
          readDatasets(id: $id, datasets: $datasets, expiresInSec: $expires_in_sec){
            __typename
            ... on SignedUrl {
              url
              file
              fields {
                name
                value
              }
            }
            ... on SignedUrls {
              urls {
                url
                file
                fields {
                  name
                  value
                }
              }
            }
          }
        }
        """
        resp = await mock_app.schema.execute(query, variable_values={"id": str(mock_pipeline.id), "datasets": [{"name": "text_in"}, {"name": "text_out"}], "expires_in_sec": 3600})
        assert resp.data["readDatasets"] is not None
        assert len(resp.data["readDatasets"]) == 2
        assert resp.data["readDatasets"][0].get("url", False)
        assert resp.data["readDatasets"][0].get("fields", False)
        assert resp.data["readDatasets"][0].get("file", False)
        assert resp.errors is None



# ui plugin examples

import plotly.graph_objects as go
import numpy as np
import pandas as pd
import tempfile
from tempfile import _TemporaryFileWrapper
import param
import panel as pn
from kedro_graphql.models import Pipeline, DataSetInput
from kedro_graphql.ui.decorators import ui_data, ui_form, ui_dashboard
from kedro_graphql.ui.components.pipeline_monitor import PipelineMonitor
from kedro_graphql.models import PipelineInput
import json
import requests
from io import StringIO

pn.extension('filedropper')


# pipeline: example00

class BaseExample00Form(pn.viewable.Viewer):
    """Base class for example00 pipeline forms.
    This class provides the basic functionality for uploading files, running the pipeline,
    and navigating to the pipeline dashboard.

    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        upload (dict): A dictionary to hold the uploaded file.
        duration (int): The duration parameter for the pipeline.
        example (str): An example string parameter for the pipeline.
        button_disabled (bool): A flag to disable the run button until a file is uploaded.

    Methods:
        navigate(pipeline_id: str): Navigate to the pipeline dashboard with the given ID.
        upload(file_dropper): Write the contents of the uploaded file to a temporary file.
        pipeline_input(): Create a PipelineInput object with the current parameters.
        run(event): Run the pipeline with the current input and parameters.
        __panel__(): Return a Panel component for the form.

    This class should be subclassed to implement specific forms for the example00 pipeline.
    """
    spec = param.Dict(default={})
    upload = param.Dict(default={})
    duration = param.Number(default=3)
    example = param.String(default="hello")
    disabled = param.Boolean(default=True)

    def navigate(self, pipeline_id: str):
        """Navigate to the pipeline dashboard with the given ID."""
        pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + "dashboard"
        pn.state.location.search = "?pipeline=example00&id=" + pipeline_id
        pn.state.location.reload = True

    async def enable(self, file_dropper):
        """
        Enable the run button by setting the disabled flag to False.
        """
        self.upload = file_dropper
        self.disabled = False

    @param.depends("upload", "duration", "example")
    async def pipeline_input(self):
        """Create a PipelineInput object with the current parameters."""

        input_dict = {"type": "text.TextDataset",
                      "filepath": "./data/01_raw/text_in.txt"}
        output_dict = {"type": "text.TextDataset",
                       "filepath": "./data/02_intermediate/text_out.txt"}

        # PipelineInput object
        return PipelineInput(**{
            "name": "example00",
            "state": "STAGED",
            "data_catalog": [
                {"name": "text_in", "config": json.dumps(input_dict)},
                {"name": "text_out", "config": json.dumps(output_dict)}
            ],
            "parameters": [
                {"name": "example", "value": self.example},
                {"name": "duration", "value": str(self.duration), "type": "FLOAT"}
            ],
            "tags": [
                {"key": "author", "value": "opensean"},
                {"key": "package", "value": "kedro-graphql"}
            ]
        })

    @param.depends('disabled')
    async def build_run_button(self):
        button = pn.widgets.Button(
            name='Run', button_type='success', disabled=self.disabled)
        pn.bind(self.run, button, watch=True)
        yield button

    async def run(self, event):
        """Stages the pipeline, creates the datasets, then updates the pipeline to READY state."""
        p_input = await self.pipeline_input()
        # Stage the pipeline
        p = await self.spec["config"]["client"].create_pipeline(pipeline_input=p_input, unique_paths=["text_in", "text_out"])
        # Create the datasets
        signed_urls = await self.spec["config"]["client"].create_datasets(id=p.id, datasets=[DataSetInput(name="text_in")], expires_in_sec=120)

        filename, content = list(self.upload.items())[0]
        files = {
            "file": (filename, content),
        }
        for f in signed_urls[0].fields:
            files.update({f.name: (None, f.value)})
        # Upload the file to the signed URL
        r = requests.post(signed_urls[0].url, files=files)

        # update the pipeline to READY state
        p_input = p.encode(encoder="input")
        p_input.state = "READY"
        p = await self.spec["config"]["client"].update_pipeline(id=p.id, pipeline_input=p_input)
        # Navigate to the pipeline dashboard
        self.navigate(p.id)

    def __panel__(self):
        raise NotImplementedError


@ui_form(pipeline="example00")
class Example00PipelineFormV1(BaseExample00Form):
    """Form for the example00 pipeline.
    This form allows users to upload a file, run the pipeline, and navigate to the pipeline dashboard.
    It inherits from BaseExample00Form and implements the __panel__ method to create the form layout.
    """

    def __init__(self, **params):
        super().__init__(**params)

    def __panel__(self):
        """Create the Panel component for the example00 pipeline form."""
        file_dropper = pn.widgets.FileDropper(multiple=False)
        # pn.bind(self.upload, file_dropper, watch=True)
        pn.bind(self.enable, file_dropper, watch=True)

        form = pn.Card(
            "An example pipeline form.",
            pn.Row(
                file_dropper
            ),
            pn.Row(
                self.build_run_button
            ),
            sizing_mode="stretch_width",
            title='Form')

        return form


@ui_form(pipeline="example00")
class Example00PipelineFormV2(BaseExample00Form):
    """Another example form for the example00 pipeline.
    This form allows users to enter additional parameters and upload a file.
    It inherits from BaseExample00Form and implements the __panel__ method to create the form layout.
    """

    def __init__(self, **params):
        super().__init__(**params)
        self.text_out = tempfile.NamedTemporaryFile(delete=False)

    def __panel__(self):
        """Create the Panel component for the example00 pipeline form with additional parameters."""
        file_dropper = pn.widgets.FileDropper(multiple=False)
        # pn.bind(self.upload, file_dropper, watch=True)
        pn.bind(self.enable, file_dropper, watch=True)

        form = pn.Card(
            "Another example pipeline form.",
            pn.Row(
                pn.widgets.TextInput.from_param(
                    self.param.example, name='example', placeholder='Enter a string here...'),
            ),
            pn.Row(
                pn.widgets.NumberInput.from_param(
                    self.param.duration, name='duration', placeholder='Enter a number here...')
            ),
            pn.Row(
                file_dropper
            ),
            pn.Row(
                self.build_run_button
            ),
            sizing_mode="stretch_width",
            title='Form')

        return form


pn.extension('echarts')


@ui_data(pipeline="example00")
class Example00Data00(pn.viewable.Viewer):
    """Data viewer for the example00 pipeline.
    It inherits from pn.viewable.Viewer and implements the __panel__ method to create the data view.

    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        id (str): The ID of the data viewer.
        pipeline (Pipeline): The Kedro pipeline associated with this data viewer.
        title (str): The title of the data viewer.
    """
    spec = param.Dict(default={})
    id = param.String(default="")
    pipeline = param.ClassSelector(class_=Pipeline)
    title = param.String(default="Plot 1")

    async def build_plot(self):
        echart_bar = {
            'title': {
                'text': 'ECharts entry example'
            },
            'tooltip': {},
            'legend': {
                'data': ['Sales']
            },
            'xAxis': {
                'data': ["shirt", "cardign", "chiffon shirt", "pants", "heels", "socks"]
            },
            'yAxis': {},
            'series': [{
                'name': 'Sales',
                'type': 'bar',
                'data': [5, 20, 36, 10, 10, 20]
            }],
        }
        yield pn.pane.ECharts(echart_bar, height=480, width=640)

    def __panel__(self):
        """Create the Panel component for the example00 data viewer."""
        pn.state.location.sync(self, {"id": "id"})
        return self.build_plot


@ui_data(pipeline="example00")
class Example00Data01(pn.viewable.Viewer):
    """Another data viewer for the example00 pipeline.
    This viewer displays a sample plot using Bokeh figures.
    It inherits from pn.viewable.Viewer and implements the __panel__ method to create the plot view.

    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        id (str): The ID of the data viewer.
        pipeline (Pipeline): The Kedro pipeline associated with this data viewer.
        title (str): The title of the data viewer.
    """
    spec = param.Dict(default={})
    id = param.String(default="")
    pipeline = param.ClassSelector(class_=Pipeline)
    title = param.String(default="Plot 1")

    async def build_plot(self):
        """Create a sample gauge plot using Panel and ECharts."""

        gauge = {
            'tooltip': {
                'formatter': '{a} <br/>{b} : {c}%'
            },
            'series': [
                {
                    'name': 'Gauge',
                    'type': 'gauge',
                    'detail': {'formatter': '{value}%'},
                    'data': [{'value': 50, 'name': 'Value'}]
                }
            ]
        }
        gauge_pane = pn.pane.ECharts(gauge, width=400, height=400)

        slider = pn.widgets.IntSlider(value=50, start=0, end=100)

        slider.jscallback(args={'gauge': gauge_pane}, value="""
        gauge.data.series[0].data[0].value = cb_obj.value
        gauge.properties.data.change.emit()
        """)

        yield pn.Column(slider, gauge_pane)

    def __panel__(self):

        return self.build_plot


# pipeline: example01
class BaseExample01Form(pn.viewable.Viewer):
    """Base class for example01 pipeline forms.
    This class provides the basic functionality for uploading files, running the pipeline,
    and navigating to the pipeline dashboard.

    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        text_in (_TemporaryFileWrapper): A temporary file wrapper for the input text file.
        uppercase (_TemporaryFileWrapper): A temporary file wrapper for the uppercase output file.
        reversed (_TemporaryFileWrapper): A temporary file wrapper for the reversed output file.
        timestamped (_TemporaryFileWrapper): A temporary file wrapper for the timestamped output file.
        duration (int): The duration parameter for the pipeline.
        example (str): An example string parameter for the pipeline.

    Methods:
        navigate(pipeline_id: str): Navigate to the pipeline dashboard with the given ID.
        upload(file_dropper): Write the contents of the uploaded file to temporary files.
        pipeline_input(): Create a PipelineInput object with the current parameters.
        run(event): Run the pipeline with the current input and parameters.
        __panel__(): Return a Panel component for the form.

    This class should be subclassed to implement specific forms for the example01 pipeline.
    """

    spec = param.Dict(default={})
    upload = param.Dict(default={})
    duration = param.Number(default=3)
    example = param.String(default="hello")
    disabled = param.Boolean(default=True)

    def navigate(self, pipeline_id: str):
        """Navigate to the pipeline dashboard with the given ID."""
        pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + "dashboard"
        pn.state.location.search = "?pipeline=example00&id=" + pipeline_id
        pn.state.location.reload = True

    async def enable(self, file_dropper):
        """
        Enable the run button by setting the disabled flag to False.
        """
        self.upload = file_dropper
        self.disabled = False

    @param.depends("upload", "duration", "example")
    async def pipeline_input(self):
        """Create a PipelineInput object with the current parameters."""

        text_in_dict = {"type": "text.TextDataset",
                        "filepath": "./data/01_raw/text_in.txt"}
        uppercase_dict = {"type": "text.TextDataset",
                          "filepath": "./data/02_intermediate/uppercased.txt"}
        reversed_dict = {"type": "text.TextDataset",
                         "filepath": "./data/02_intermediate/reversed.txt"}
        timestamped_dict = {"type": "text.TextDataset",
                            "filepath": "./data/02_intermediate/timestamped.txt"}
        timestamped_partitioned_dict = {"type": "partitions.PartitionedDataset",
                                        "path": "./data/02_intermediate/timestamped_partitioned",
                                        "filename_suffix": ".txt",
                                        "dataset": {"type": "text.TextDataset"}}

        # PipelineInput object
        return PipelineInput(**{
            "name": "example01",
            "state": "STAGED",
            "data_catalog": [{"name": "text_in", "config": json.dumps(text_in_dict)},
                             {"name": "uppercased",
                                 "config": json.dumps(uppercase_dict)},
                             {"name": "reversed", "config": json.dumps(reversed_dict)},
                             {"name": "timestamped",
                                 "config": json.dumps(timestamped_dict)},
                             {"name": "timestamped_partitioned",
                                 "config": json.dumps(timestamped_partitioned_dict)}
                             ],
            "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
        })

    @param.depends('disabled')
    async def build_run_button(self):
        button = pn.widgets.Button(
            name='Run', button_type='success', disabled=self.disabled)
        pn.bind(self.run, button, watch=True)
        yield button

    async def run(self, event):
        """Stages the pipeline, creates the datasets, then updates the pipeline to READY state."""
        p_input = await self.pipeline_input()
        # Stage the pipeline
        p = await self.spec["config"]["client"].create_pipeline(pipeline_input=p_input, unique_paths=["text_in", "uppercased", "reversed", "timestamped", "timestamped_partitioned"])
        # Create the datasets
        signed_urls = await self.spec["config"]["client"].create_datasets(id=p.id, datasets=[DataSetInput(name="text_in")], expires_in_sec=120)

        filename, content = list(self.upload.items())[0]
        files = {
            "file": (filename, content),
        }
        for f in signed_urls[0].fields:
            files.update({f.name: (None, f.value)})
        # Upload the file to the signed URL
        r = requests.post(signed_urls[0].url, files=files)

        # update the pipeline to READY state
        p_input = p.encode(encoder="input")
        p_input.state = "READY"
        p = await self.spec["config"]["client"].update_pipeline(id=p.id, pipeline_input=p_input)
        # Navigate to the pipeline dashboard
        self.navigate(p.id)

    def __panel__(self):
        raise NotImplementedError


@ui_form(pipeline="example01")
class Example01PipelineFormV1(BaseExample01Form):
    """Form for the example01 pipeline.
    This form allows users to upload a file, run the pipeline, and navigate to the pipeline dashboard.
    It inherits from BaseExample01Form and implements the __panel__ method to create the form layout.
    """

    def __init__(self, **params):
        super().__init__(**params)

    def __panel__(self):
        file_dropper = pn.widgets.FileDropper(multiple=False)
        pn.bind(self.enable, file_dropper, watch=True)

        form = pn.Card(
            "An example pipeline form.",
            pn.Row(
                file_dropper
            ),
            pn.Row(
                self.build_run_button
            ),
            sizing_mode="stretch_width",
            title='Form')

        return form


pn.extension('plotly')


@ui_dashboard(pipeline="example01")
class Example01PipelineUIV1(pn.viewable.Viewer):
    """Dashboard for the example01 pipeline.
    This dashboard displays the pipeline stages and allows users to monitor the pipeline's progress.
    It inherits from pn.viewable.Viewer and implements the __panel__ method to create the dashboard layout.
    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        id (str): The ID of the pipeline.
        pipeline (Pipeline): The Kedro pipeline associated with this dashboard.
    """

    spec = param.Dict(default={})
    id = param.String(default="")
    pipeline = param.ClassSelector(class_=Pipeline)

    def __init__(self, **params):
        super().__init__(**params)

    def draw_pipeline(self):
        nodes = ["stage 1", "stage 2", "stage 3"]
        # it would be nice to get nodes from the pipeline object instead

        # Define node colors, default color is blue
        node_colors = ['blue'] * len(nodes)
        node_colors[1] = 'green'  # Change color of the second node to green

        # Define edges as tuples of node indices
        node_trace = go.Scatter(
            x=nodes,
            y=[0, 0, 0],
            mode='lines+markers',
            text=[nodes],
            marker=dict(
                size=20,
                color=node_colors
            ),
            line=dict(width=2, color='gray'),
            textposition='bottom center'
        )
        fig = go.Figure(data=[node_trace],
                        layout=go.Layout(showlegend=False))

        fig.update_layout({
            "plot_bgcolor": "rgba(0, 0, 0, 0)",
            "paper_bgcolor": "rgba(0, 0, 0, 0)",
            'xaxis': {'showgrid': False, 'zeroline': False, 'showticklabels': True},
            'yaxis': {'showgrid': False, 'zeroline': False, 'showticklabels': False},
            'hovermode': False,
            'height': 100,
            'margin': dict(t=0, b=0, l=0, r=0),
        })
        return pn.pane.Plotly(fig, config={'displayModeBar': False, 'scrollZoom': False})

    @param.depends("spec", "pipeline")
    async def build_ui(self):
        yield pn.indicators.LoadingSpinner(value=True, width=25, height=25)
        monitor = PipelineMonitor(
            client=self.spec["config"]["client"], pipeline=self.pipeline)
        ui = pn.Column(
            pn.Row(self.draw_pipeline()),
            pn.Row(monitor),
            sizing_mode="stretch_width")

        yield ui

    def __panel__(self):

        pn.state.location.sync(self, {"id": "id"})
        return self.build_ui



import json
from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional
# from kedro.io.core import _parse_filepath
# from .signed_url.local_file_provider import LocalFileProvider
# from .signed_url.base import PreSignedUrlProvider
# from importlib import import_module
import strawberry
from bson.objectid import ObjectId
from fastapi.encoders import jsonable_encoder
from kedro.io import AbstractDataset
from kedro.io.core import _parse_filepath
from strawberry.utils.str_converters import to_camel_case, to_snake_case
from cloudevents.conversion import to_json
from cloudevents.pydantic.v1 import CloudEvent
from pathlib import Path

from kedro_graphql.exceptions import DataSetConfigError
# from strawberry.permission import PermissionExtension

from .config import load_config
from .logs.logger import logger
# from .permissions import get_permissions


CONFIG = load_config()
# logger.debug("configuration loaded by {s}".format(s=__name__))
##
# PERMISSIONS_CLASS = get_permissions(CONFIG.get("KEDRO_GRAPHQL_PERMISSIONS"))
# logger.info("{s} using permissions class: {d}".format(s=__name__, d=PERMISSIONS_CLASS))


def mark_deprecated(default=None):
    return strawberry.field(default=default, deprecation_reason="see " + str(CONFIG["KEDRO_GRAPHQL_DEPRECATIONS_DOCS"]))


@strawberry.type
class Tag:
    key: str
    value: str


@strawberry.input
class TagInput:
    key: str
    value: str


@strawberry.enum
class ParameterType(Enum):
    STRING = "string"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    FLOAT = "float"


@strawberry.type
class Parameter:
    name: str
    value: str
    type: Optional[ParameterType] = ParameterType.STRING

    @staticmethod
    def decode(input_dict) -> dict:
        """
        Returns a Parameter object from a dictionary.
        """
        if input_dict.get("type", False):
            return Parameter(
                name=input_dict["name"],
                value=input_dict["value"],
                type=ParameterType[input_dict["type"].upper()])
        else:
            return Parameter(**input_dict)

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        value = self.value
        if self.type == "boolean":
            value = self.value.lower()
            if value == "true":
                value = True
            elif value == "false":
                value = False
            else:
                raise ValueError(
                    "Parameter of type BOOL must be one of 'True', 'true', 'False', or 'false'")

        elif self.type == "integer":
            value = int(self.value)

        elif self.type == "float":
            value = float(self.value)

        return {self.name: value}


@strawberry.input
class ParameterInput:
    name: str
    value: str
    type: Optional[ParameterType] = ParameterType.STRING

    @staticmethod
    def create(parameters: dict):

        def _create_parameter_input(name, value):

            type_map = [
                (int, ParameterType.INTEGER),
                (float, ParameterType.FLOAT),
                (bool, ParameterType.BOOLEAN),
                (str, ParameterType.STRING),
            ]

            try:
                type_enum: ParameterType = next(
                    t[1] for t in type_map if type(value) == t[0])
            except StopIteration:
                raise ValueError(
                    f"Only primitive types are supported ({' '.join(str(x[0]) for x in type_map)}). Got {type(value)}"
                )

            return ParameterInput(name=name, value=str(value), type=type_enum.value.upper())

        params = [_create_parameter_input(k, v) for k, v in parameters.items()]
        return params


@strawberry.input
class CredentialSetInput:
    name: str
    value: str

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        return {self.name: self.value}


@strawberry.input
class CredentialInput:
    name: str
    value: List[CredentialSetInput]

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        values = {}
        for v in self.value:
            values.update(v.serialize())
        return {self.name: values}


@strawberry.input
class CredentialNestedInput:
    name: str
    value: List[CredentialInput]

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        values = {}
        for v in self.value:
            values.update(v.serialize())
        return {self.name: values}


@strawberry.type
class DataSet:
    name: str
    config: Optional[str] = None
    # credentials: Optional[List[CredentialInput]]
    tags: Optional[List[Tag]] = None

    @strawberry.field
    def exists(self) -> bool:
        if self.config:
            return AbstractDataset.from_config(self.name, self.parse_config()).exists()
        else:
            return False

    @strawberry.field
    def partitions(self) -> Optional[List[str]]:
        config = self.parse_config()
        if not config.get("type", None):
            raise DataSetConfigError(
                "Invalid dataset configuration. Must have 'type' key")
        elif config["type"] != "partitions.PartitionedDataset":
            raise DataSetConfigError(
                "Dataset is not a PartitionedDataset. 'partitions' field is only available for PartitionedDatasets."
            )
        else:
            partitions = AbstractDataset.from_config(self.name, config).load()
            return list(partitions.keys())

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        temp = self.__dict__.copy()
        temp.pop("name")
        return {self.name: json.loads(temp['config'])}

    @staticmethod
    def decode(payload):
        """
        Return a new DataSet from a dictionary.

        Args:
            payload (dict): dict representing DataSet e.g.

                {
                  "name": "text_in",
                  "config": '{"filepath": "./data/01_raw/text_in.txt", "type": "text.TextDataSet", "save_args": [{"name": "say", "value": "hello"}], "load_args": [{"name": "say", "value": "hello"}]}',
                  "tags":[{"key": "owner name", "value": "harinlee0803"},{"key": "owner email", "value": "test@example.com"}]
                }

        """
        if payload.get("tags", False):
            tags = [Tag(**t) for t in payload["tags"]]
        else:
            tags = None

        return DataSet(
            name=payload["name"],
            config=payload["config"],
            tags=tags
        )

    def parse_config(self) -> dict:
        """
        Return the config as a dictionary.

        Example usage:

            from kedro_graphql.models import DataSet

            d = DataSet(name="text_in", config='{"filepath": "./data/01_raw/text_in.txt", "type": "text.TextDataSet", "save_args": [{"name": "say", "value": "hello"}], "load_args": [{"name": "say", "value": "hello"}]}')

            print(d.parse_config())

            {'filepath': './data/01_raw/text_in.txt',
             'type': 'text.TextDataSet',
             'save_args': [{'name': 'say', 'value': 'hello'}],
             'load_args': [{'name': 'say', 'value': 'hello'}]}

        """
        try:
            return json.loads(self.config)
        except json.JSONDecodeError as e:
            raise DataSetConfigError(f"Unable to parse JSON in config: {e}")
        except Exception as e:
            raise DataSetConfigError(f"Invalid dataset configuration: {e}")

    def parse_filepath(self) -> tuple[str, str]:
        """
        Parse the filepath from the dataset configuration.

        Args:
            config (dict): The dataset configuration.

        Returns:
            tuple[str, str]: (protocol, the file path).
        """
        config = self.parse_config()
        filepath = config.get("filepath", None)
        if not filepath:
            raise DataSetConfigError(
                "Invalid dataset configuration. Must have 'filepath' key")

        return _parse_filepath(filepath)["protocol"], filepath

    def parse_path(self) -> tuple[str, str]:
        """
        Parse the path from the dataset configuration.

        Args:
            config (dict): The dataset configuration.

        Returns:
            tuple[str, str]: (protocol, the file path).
        """
        path = self.parse_config().get("path", None)
        if not path:
            raise DataSetConfigError(
                "Invalid dataset configuration. Must have 'path' key")

        return _parse_filepath(path)["protocol"], path


@strawberry.input
class DataSetInput:
    name: str
    config: Optional[str] = None
    tags: Optional[List[TagInput]] = None
    partitions: Optional[List[str]] = None

    def encode(self, encoder="graphql"):
        if encoder == "dict":
            return jsonable_encoder(self)
        elif encoder == "graphql":
            p = jsonable_encoder(self)
            p = {to_camel_case(k): v for k, v in p.items()}
            return p
        else:
            raise TypeError("encoder must be 'dict' or 'graphql'")


class DataCatalog:
    datasets: List[DataSet]


@strawberry.input
class DataCatalogInput:
    datasets: List[DataSetInput]

    @staticmethod
    def create(config):
        """
        context.config_loader["catalog"]

        {'text_in': {'type': 'text.TextDataSet',
                     'filepath': './data/01_raw/text_in.txt'},
         'text_out': {'type': 'text.TextDataSet',
                      'filepath': './data/02_intermediate/text_out.txt'}}

        Example usage:

            from kedro_graphql.models import DataCatalogInput

            catalog = DataCatalogInput.create(context.config_loader["catalog"])

            print(catalog)

            [DataSetInput(name='text_in', config='{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}', type=None, filepath=None, save_args=None, load_args=None, credentials=None),
             DataSetInput(name='text_out', config='{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}', type=None, filepath=None, save_args=None, load_args=None, credentials=None)]

        """
        return [DataSetInput(name=k, config=json.dumps(v)) for k, v in config.items()]


@strawberry.type
class Node:
    name: str
    inputs: List[str]
    outputs: List[str]
    tags: List[str]


@strawberry.type(description="PipelineTemplates are definitions of Pipelines.  They represent the supported interface for executing a Pipeline.")
class PipelineTemplate:
    id: str = strawberry.field(description="ID of the pipeline template.")
    name: str
    kedro_pipelines: strawberry.Private[dict]
    kedro_catalog: strawberry.Private[dict]
    kedro_parameters: strawberry.Private[dict]

    @strawberry.field
    def describe(self) -> str:
        return self.kedro_pipelines[self.name].describe()

    @strawberry.field
    def nodes(self) -> List[Node]:
        nodes = self.kedro_pipelines[self.name].nodes

        return [Node(name=n.name, inputs=n.inputs, outputs=n.outputs, tags=n.tags) for n in nodes]

    @strawberry.field
    def parameters(self) -> List[Parameter]:
        # keep track of parameters to avoid returning duplicates
        params = {}
        for n in self.kedro_pipelines[self.name].all_inputs():
            if n.startswith("params:"):
                name = n.split("params:")[1]
                value = self.kedro_parameters[name]
                if not params.get(name, False):
                    params[name] = value
            elif n == "parameters":
                for k, v in self.kedro_parameters.items():
                    if not params.get(k, False):
                        params[k] = v
        return [Parameter(name=k, value=v) for k, v in params.items()]

    @strawberry.field
    def inputs(self) -> List[DataSet]:
        inputs_resolved = []
        for n in self.kedro_pipelines[self.name].all_inputs():
            if not n.startswith("params:") and n != "parameters":
                config = self.kedro_catalog[n]
                inputs_resolved.append(DataSet(name=n, config=json.dumps(config)))

        return inputs_resolved

    @strawberry.field
    def outputs(self) -> List[DataSet]:
        outputs_resolved = []
        for n in self.kedro_pipelines[self.name].all_outputs():
            if self.kedro_catalog.get(n, None):
                config = self.kedro_catalog[n]
                outputs_resolved.append(DataSet(name=n, config=json.dumps(config)))
            else:
                logger.warning(
                    f"PipelineTemplate '{self.name}' has an output '{n}' that is not defined in the catalog. This may be due to a missing dataset configuration or because it is a MemoryDataset."
                )

        return outputs_resolved


@strawberry.type
class PageMeta:
    next_cursor: Optional[str] = strawberry.field(
        description="The next cursor to continue with."
    )


@strawberry.type
class PipelineTemplates:
    pipeline_templates: List[PipelineTemplate] = strawberry.field(
        description="The list of pipeline templates.")

    page_meta: PageMeta = strawberry.field(description="Metadata to aid in pagination.")

    @staticmethod
    def _build_pipeline_index(kedro_pipelines, kedro_catalog, kedro_parameters):
        """
        """
        pipes = []
        count = 100000000000000000000000
        for k, v in kedro_pipelines.items():
            pipes.append(PipelineTemplate(name=k,
                                          id=ObjectId(str(count)),
                                          kedro_pipelines=kedro_pipelines,
                                          kedro_catalog=kedro_catalog,
                                          kedro_parameters=kedro_parameters))
            count += 1

        return pipes


@strawberry.enum
class PipelineSliceType(Enum):
    TAGS = "tags"
    FROM_NODES = "from_nodes"
    TO_NODES = "to_nodes"
    NODE_NAMES = "node_names"
    FROM_INPUTS = "from_inputs"
    TO_OUTPUTS = "to_outputs"
    NODE_NAMESPACE = "node_namespace"


@strawberry.input(description="Slice a pipeline.")
class PipelineSlice:
    slice: PipelineSliceType
    args: List[str]  # e.g. ["node1", "node2"]


@strawberry.enum
class PipelineInputStatus(Enum):
    STAGED = "STAGED"
    READY = "READY"


@strawberry.input(description="PipelineInput")
class PipelineInput:
    name: str
    state: PipelineInputStatus = PipelineInputStatus.STAGED
    parameters: Optional[List[ParameterInput]] = None
    data_catalog: Optional[List[DataSetInput]] = None
    tags: Optional[List[TagInput]] = None
    parent: Optional[strawberry.ID] = None
    runner: Optional[str] = None
    slices: Optional[List[PipelineSlice]] = None
    only_missing: Optional[bool] = False

    @staticmethod
    def create(name=None, data_catalog=None, parameters=None, tags=None):
        """
        Example usage:

            from kedro_graphql.models import PipelineInput
            from fastapi.encoders import jsonable_encoder

            p = PipelineInput(name = "example00",
                         data_catalog = context.config_loader["catalog"],
                         parameters = context.config_loader["parameters"],
                         tags = [{""owner":"person"}])

            print(p)

            PipelineInput(name='example00',
                          parameters=[
                            ParameterInput(name='example',
                                           value='hello',
                                           type=<ParameterType.STRING: 'string'>),
                            ParameterInput(name='duration', value='1', type=<ParameterType.STRING: 'string'>)
                          ],
                          data_catalog=[
                            DataSetInput(
                                name='text_in', config='{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}'),
                            DataSetInput(
                                name='text_out', config='{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}')
                          ],
                          tags=[TagInput(key='owner', value='sean')])

            print(jsonable_encoder(p))

            ## this can be used as the PipelineInput parameter when calleing the pipeline mutation via the API
            {'name': 'example00',
            'parameters': [{'name': 'example', 'value': 'hello', 'type': 'string'},
             {'name': 'duration', 'value': '1', 'type': 'string'}],
            'data_catalog': [{'name': 'text_in',
              'config': '{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}'},
             {'name': 'text_out',
              'config': '{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}'}],
            'tags': [{'key': 'owner', 'value': 'sean'}],
            'credentials': None,
            'credentials_nested': None}

        """
        if tags:
            tags = [TagInput(key=k, value=v) for t in tags for k, v in tags.items()]

        if data_catalog:
            data_catalog = DataCatalogInput.create(data_catalog)

        if parameters:
            parameters = ParameterInput.create(parameters)

        return PipelineInput(name=name,
                             parameters=parameters,
                             data_catalog=data_catalog,
                             tags=tags)

    def encode(self, encoder="graphql"):
        if encoder == "dict":
            return jsonable_encoder(self)
        elif encoder == "graphql":
            p = jsonable_encoder(self)
            p = {to_camel_case(k): v for k, v in p.items()}
            # make sure parameter types are uppercase
            if p.get("parameters", None):
                for param in p["parameters"]:
                    if param.get("type", None):
                        param["type"] = param["type"].upper()
            return p
        else:
            raise TypeError("encoder must be 'dict' or 'graphql'")

    @classmethod
    def from_event(cls, name: str, state: PipelineInputStatus, event: CloudEvent) -> "PipelineInput":
        """
        Factory method to create a new PipelineInput from a CloudEvent. 
        Tags will be added for event metadata and the entire event will be added 
        as a single parameter (json-serialized).
        """
        event_bytes = to_json(event)
        event = json.loads(event_bytes.decode())

        id = event.get("id")
        source = event.get("source")
        type = event.get("type")

        if not id or not source or not type:
            raise ValueError(
                "CloudEvent must have 'id', 'source', and 'type' attributes")

        return cls(
            name=name,
            state=state,
            parameters=[ParameterInput(
                name="event", value=event_bytes, type="STRING")],
            tags=[
                {"key": "event_id", "value": id},
                {"key": "event_source", "value": source},
                {"key": "event_type", "value": type}
            ]
        )


@strawberry.enum
class State(Enum):
    READY = 'READY'
    STAGED = 'STAGED'
    STARTED = 'STARTED'
    RETRY = 'RETRY'
    FAILURE = 'FAILURE'
    SUCCESS = 'SUCCESS'
    REVOKED = 'REVOKED'
    PENDING = 'PENDING'
    RECIEVED = 'RECIEVED'


@strawberry.type
class PipelineStatus:
    state: State
    session: Optional[str]
    runner: Optional[str] = None
    filtered_nodes: Optional[List[str]] = None
    started_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    task_id: Optional[str] = None
    task_name: Optional[str] = None
    task_args: Optional[str] = None
    task_kwargs: Optional[str] = None
    task_request: Optional[str] = None
    task_exception: Optional[str] = None
    task_traceback: Optional[str] = None
    task_einfo: Optional[str] = None
    task_result: Optional[str] = None


@strawberry.type
class Pipeline:
    id: Optional[strawberry.ID] = None
    name: str
    data_catalog: Optional[List[DataSet]] = None
    describe: Optional[str] = None
    nodes: Optional[List[Node]] = None
    parameters: Optional[List[Parameter]] = None
    status: List[PipelineStatus] = strawberry.field(default_factory=list)
    tags: Optional[List[Tag]] = None
    created_at: Optional[datetime] = None
    parent: Optional[strawberry.ID] = None
    project_version: Optional[str] = None
    pipeline_version: Optional[str] = None
    kedro_graphql_version: Optional[str] = None

    def serialize(self):
        parameters = {}
        data_catalog = {}

        if self.parameters:
            for p in self.parameters:
                s = p.serialize()
                parameters.update(s)

        if self.data_catalog:
            for d in self.data_catalog:
                s = d.serialize()
                data_catalog.update(s)

        return {
            "id": str(self.id),
            "name": self.name,
            "data_catalog": data_catalog,
            "parameters": parameters,
        }

    def encode(self, encoder="dict"):

        if encoder == "dict":
            p = deepcopy(self)
            # if type ObjectID the jsonable_encoder will throw an error
            p.id = str(p.id)
            encoded_pipeline = jsonable_encoder(p)

            return encoded_pipeline
        elif encoder == "kedro":
            return self.serialize()
        elif encoder == "input":
            if self.parameters:
                parameters = [ParameterInput(name=p.name, value=p.value, type=p.type.value)
                              for p in self.parameters]
            else:
                parameters = None
            return PipelineInput(
                name=self.name,
                data_catalog=[DataSetInput(name=d.name, config=d.config)
                              for d in self.data_catalog],
                parameters=parameters,
                tags=[TagInput(key=t.key, value=t.value)
                      for t in self.tags] if self.tags else None
            )
        else:
            raise TypeError("encoder must be 'dict', 'kedro', or 'input'")

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new Pipeline from a dictionary or graphql api response.
        """
        if decoder == "graphql":
            payload = {to_snake_case(k): v for k, v in payload.items()}
            if payload["status"]:
                payload["status"] = [
                    {to_snake_case(k): v for k, v in s.items()} for s in payload["status"]]

            return cls.decode_dict(payload)

        elif decoder == "dict" or isinstance(payload, dict):
            return cls.decode_dict(payload)

        elif isinstance(payload, PipelineInput):
            return cls.decode_pipeline_input(payload)

    @staticmethod
    def decode_dict(payload):
        if payload.get("tags", None):
            tags = [Tag(**t) for t in payload["tags"]]
        else:
            tags = None

        if payload.get("data_catalog", None):
            data_catalog = [DataSet.decode(d) for d in payload["data_catalog"]]
        else:
            data_catalog = []

        if payload.get("status", None):
            status = [PipelineStatus(
                state=State[s["state"]],
                session=s["session"],
                runner=s.get("runner", "kedro.runner.SequentialRunner"),
                filtered_nodes=s.get("filtered_nodes"),
                started_at=datetime.fromisoformat(
                    s["started_at"]) if s.get("started_at") else None,
                finished_at=datetime.fromisoformat(
                    s["finished_at"]) if s.get("finished_at") else None,
                task_name=s.get("task_name"),
                task_id=s.get("task_id"),
                task_args=s.get("task_args"),
                task_kwargs=s.get("task_kwargs"),
                task_request=s.get("task_request"),
                task_exception=s.get("task_exception"),
                task_traceback=s.get("task_traceback"),
                task_einfo=s.get("task_einfo"),
                task_result=s.get("task_result")
            ) for s in payload["status"]]
        else:
            status = []

        if payload.get("parameters", None):
            parameters = [Parameter.decode(p) for p in payload["parameters"]]
        else:
            parameters = None

        return Pipeline(
            id=payload.get("id", None),
            name=payload["name"],
            data_catalog=data_catalog,
            parameters=parameters,
            status=status,
            tags=tags,
            created_at=datetime.fromisoformat(
                payload["created_at"]) if payload.get("created_at", None) else None,
            parent=payload.get("parent", None),
            project_version=payload.get("project_version", None),
            pipeline_version=payload.get("pipeline_version", None),
            kedro_graphql_version=payload.get("kedro_graphql_version", None)
        )

    @classmethod
    def decode_pipeline_input(cls, payload):
        """Factory method to create a new Pipeline from a PipelineInput object.
        """
        d = jsonable_encoder(payload)
        return cls.decode_dict(d)


@strawberry.type
class Pipelines:
    pipelines: List[Pipeline] = strawberry.field(
        description="The list of pipeline instances.")
    page_meta: PageMeta = strawberry.field(description="Metadata to aid in pagination.")

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new Pipelines from a graphql api response.
        """
        if decoder == "graphql":
            meta = {to_snake_case(k): v for k,
                    v in payload["readPipelines"]["pageMeta"].items()}
            return Pipelines(page_meta=PageMeta(**meta),
                             pipelines=[Pipeline.decode(p, decoder="graphql") for p in payload["readPipelines"]["pipelines"]])
        else:
            raise TypeError("decoder must be 'graphql'")


@strawberry.type
class PipelineEvent:
    id: str
    task_id: str
    status: str
    result: Optional[str] = None
    timestamp: str
    traceback: Optional[str] = None

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new PipelineEvent from a graphql api response.
        """
        if decoder == "graphql":
            result = {to_snake_case(k): v for k, v in payload["pipeline"].items()}
            return PipelineEvent(**result)
        else:
            raise TypeError("decoder must be 'graphql'")


@strawberry.type
class PipelineLogMessage:
    id: str
    message: str
    message_id: str
    task_id: str
    time: str

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new PipelineLogMessage from a graphql api response.
        """
        if decoder == "graphql":
            result = {to_snake_case(k): v for k, v in payload["pipelineLogs"].items()}
            return PipelineLogMessage(id=result["id"],
                                      message=result.get("message", ""),
                                      message_id=result.get("message_id", ""),
                                      task_id=result.get("task_id", ""),
                                      time=result.get("time", ""))
        else:
            raise TypeError("decoder must be 'graphql'")


@strawberry.type
class SignedUrlField:
    name: Optional[str] = None
    value: Optional[str] = None


@strawberry.type
class SignedUrl:
    url: str
    file: str
    fields: Optional[List[SignedUrlField]] = None

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new SignedUrl from a graphql api response.
        """
        if decoder == "graphql":
            result = {to_snake_case(k): v for k, v in payload.items()}
            return SignedUrl(url=result["url"], file=result["file"], fields=[SignedUrlField(name=f["name"], value=f["value"]) for f in result.get("fields", None)])
        else:
            raise TypeError("decoder must be 'graphql'")

    def get_field_value(self, name: str) -> str | None:
        """
        Extract a field value from an array of SignedUrlFields.

        Args:
            name (str): The name of the field to extract.

        Returns:
            str | None: The value of the field, or None if not found.
        """
        for f in self.fields:
            if f.name == name:
                return f.value


@strawberry.type
class SignedUrls:
    urls: List[SignedUrl]

    @classmethod
    def decode(cls, payload, decoder=None):
        """Factory method to create a new SignedUrls from a graphql api response.
        """
        if decoder == "graphql":
            result = {to_snake_case(k): v for k, v in payload.items()}
            urls = [SignedUrl.decode(u, decoder="graphql") for u in result["urls"]]
            return SignedUrls(urls=urls)
        else:
            raise TypeError("decoder must be 'graphql'")



import asyncio
import panel as pn
import param
import requests
from kedro.io import AbstractDataset

pn.extension("perspective")

class DatasetPerspective(pn.viewable.Viewer):
    """
    A Panel component that loads a Kedro dataset from a presigned URL and displays it using pn.pane.Perspective.

    Attributes:
        spec (dict): Kedro GraphQL UI specification
        presigned_url (str): The presigned URL to load data from.
        ds_name (str): The name of the dataset.
        ds_type (str): The type of the dataset.
        file_size_limit_mb (int): Maximum file size limit in MB to prevent loading large datasets
    """

    spec = param.Dict(default={}, doc="Kedro GraphQL UI specification")
    presigned_url = param.String(doc="The presigned URL to load data from")
    ds_name = param.String(doc="The name of the Kedro dataset")
    ds_type = param.String(doc="The type of the Kedro dataset")
    file_size_limit_mb = param.Integer(doc="Maximum file size limit in MB")

    def __init__(self, **params):
        super().__init__(**params)

    async def build_component(self):
        yield pn.Column(
            pn.indicators.LoadingSpinner(value=True, width=25, height=25),
            pn.pane.Markdown("Fetching data... This may take some time if the dataset is large."),
            sizing_mode="stretch_width",
        )

        query = pn.state.location.query_params

        presigned_url = query.get("presigned_url")
        ds_name = query.get("ds_name")
        ds_type = query.get("ds_type")

        if not presigned_url:
            yield pn.Column(
                "# Dataset Perspective",
                pn.pane.Markdown("**Missing `presigned_url` query parameter.**"),
                sizing_mode="stretch_width",
            )

        if not ds_name:
            yield pn.Column(
                "# Dataset Perspective",
                pn.pane.Markdown("**Missing `ds_name` query parameter.**"),
                sizing_mode="stretch_width",
            )

        if not ds_type:
            yield pn.Column(
                "# Dataset Perspective",
                pn.pane.Markdown("**Missing `ds_name` or `ds_type` query parameter.**"),
                sizing_mode="stretch_width",
            )

        try:
            # Check file size before loading
            headers = {"Range": "bytes=0-0"}
            response = requests.get(presigned_url, headers=headers)
            if response.status_code in (200, 206):
                size_str = response.headers.get('Content-Range') or response.headers.get('Content-Length')
                if size_str:
                    if 'Content-Range' in response.headers:
                        # Format: bytes 0-0/123456
                        size_bytes = int(size_str.split('/')[-1])
                    else:
                        size_bytes = int(size_str)
                    max_size = self.file_size_limit_mb * 1024 * 1024  # 10 MB
                    if size_bytes > max_size:
                        yield pn.Column("# Dataset Perspective", pn.pane.Markdown(f"### File size {size_bytes/1024/1024:.2f} MB exceeds {self.file_size_limit_mb} MB limit."))
                        return

            dataset = AbstractDataset.from_config(
                name=self.ds_name,
                config={
                    "type": ds_type,
                    "filepath": presigned_url
                }
            )
            df = await asyncio.to_thread(dataset.load)

            yield pn.Column(
                "# Dataset Perspective",
                pn.pane.Perspective(
                    df, height=1000, sizing_mode="stretch_width", theme="pro-dark",
                    editable=False, settings=False),
                sizing_mode="stretch_width",
            )

        except Exception as e:
            yield pn.Column(
                "# Dataset Perspective",
                pn.pane.Markdown(f"**Error loading data from URL:** {str(e)}"),
                sizing_mode="stretch_width",
            )

    def __panel__(self):
        return self.build_component



class DataSetConfigError(Exception):
    """``DataSetConfigError`` raised when dataset configuration is invalid or cannot be parsed. 
    """

    pass


class DataSetError(Exception):
    """``DataSetError`` raised when dataset operations fail.
    """

    pass



"""
This is a boilerplate pipeline 'event00'
generated using Kedro 0.19.11
"""

from kedro.pipeline import node, Pipeline, pipeline  # noqa
from .nodes import run_example00_via_api


def create_pipeline(**kwargs) -> Pipeline:
    return pipeline(
        [
            node(run_example00_via_api,
                 inputs=["params:id", "params:event"], outputs="pipeline")
        ]
    )



import abc
from ..models import DataSet, SignedUrl, SignedUrls
from strawberry.types import Info


class SignedUrlProvider(metaclass=abc.ABCMeta):
    """
    Abstract base class for providing signed URLs for reading and creating kedro datasets
    """

    @abc.abstractmethod
    def read(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Abstract method to get a signed URL for downloading a dataset.

        Args:
            info (Info): Strawberry GraphQL Info object.
            dataset (DataSet): The dataset for which to create a signed URL.
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
            partitions (list | None): Optional list of partitions for datasets with multiple files e.g. a PartitionedDataset.

        Returns:
            str | None: A signed URL for downloading the dataset.
        """
        pass

    @abc.abstractmethod
    def create(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Abstract method to get a signed URL for uploading a dataset.

        Args:
            info (Info): Strawberry GraphQL Info object.
            dataset (DataSet): The dataset for which to create a signed URL.
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
            partitions (list | None): Optional list of partitions for datasets with multiple files e.g. a PartitionedDataset.

        Returns:
            dict | None: A dictionary with the URL to post to and form fields and values to submit with the POST.
        """
        pass



from fastapi import FastAPI, Request, Depends, HTTPException, status, HTTPException, UploadFile, File, Form
from fastapi.responses import FileResponse
import jwt
import shutil
from pathlib import Path
from strawberry.fastapi import GraphQLRouter
from cloudevents.http import from_http, to_json
from cloudevents.pydantic.v1 import CloudEvent
from contextlib import asynccontextmanager

from .logs.logger import logger
from .backends import init_backend
from .celeryapp import celery_app
from .decorators import RESOLVER_PLUGINS, TYPE_PLUGINS, discover_plugins
from .models import PipelineTemplates
from .schema import build_schema
from .config import load_config
from .permissions import get_permissions
from starlette.requests import Request
from kedro_graphql.utils import build_graphql_query
from kedro_graphql.models import PipelineInput, Pipeline, ParameterInput

CONFIG = load_config()
logger.debug("configuration loaded by {s}".format(s=__name__))

PERMISSIONS_CLASS = get_permissions(CONFIG.get("KEDRO_GRAPHQL_PERMISSIONS"))


@asynccontextmanager
async def lifespan(app: FastAPI):
    app.backend.startup()
    yield
    app.backend.shutdown()


class KedroGraphQL(FastAPI):

    def __init__(self, kedro_session=None, config=CONFIG, lifespan_handler=None):
        super(KedroGraphQL, self).__init__(
            title=config["KEDRO_GRAPHQL_APP_TITLE"],
            description=config["KEDRO_GRAPHQL_APP_DESCRIPTION"],
            version=config["KEDRO_GRAPHQL_PROJECT_VERSION"],
            docs_url="/docs",  # Swagger UI URL
            root_path=config["KEDRO_GRAPHQL_ROOT_PATH"],
            lifespan=lifespan_handler or lifespan
        )

        self.kedro_session = kedro_session
        self.kedro_context = self.kedro_session.load_context()
        self.kedro_catalog = self.kedro_context.config_loader["catalog"]
        self.kedro_parameters = self.kedro_context.config_loader["parameters"]
        from kedro.framework.project import pipelines
        self.kedro_pipelines = pipelines
        self.kedro_pipelines_index = PipelineTemplates._build_pipeline_index(
            self.kedro_pipelines, self.kedro_catalog, self.kedro_parameters)

        self.config = config

        self.resolver_plugins = RESOLVER_PLUGINS
        self.type_plugins = TYPE_PLUGINS

        discover_plugins(self.config)
        self.schema = build_schema(self.type_plugins)
        self.backend = init_backend(self.config)
        self.graphql_app = GraphQLRouter(self.schema)
        self.include_router(self.graphql_app, prefix="/graphql")
        self.add_api_websocket_route("/graphql", self.graphql_app)

        self.celery_app = celery_app(self.config, self.backend, self.schema)

        class Info:
            """A simple class to hold the request context for permissions."""

            def __init__(self, request: Request):
                self.context = {"request": request}

        @staticmethod
        def authenticate_factory(action: str = None):
            """Factory function to create an authentication dependency.

            Kwargs:
                action (str): The action for which the user needs to be authenticated.
            Returns:
                function: A function that checks if the user is authenticated.
            """

            def authenticate(request: Request):
                """Dependency to authenticate the user based on permissions.
                This function checks if the user is authenticated by verifying
                the permissions class. If the user is not authenticated, it raises
                an HTTPException with a 403 Forbidden status code.
                This function is used as a dependency in the event endpoint to ensure
                that only authenticated users can create events.

                Args:
                    request (Request): The incoming request.
                Returns:
                    bool: True if the user is authenticated, False otherwise.
                Raises:
                    HTTPException: If the user is not authenticated.
                """

                access = PERMISSIONS_CLASS(action=action).has_permission(
                    None, Info(request))
                if not access:
                    raise HTTPException(detail="User is not authenticated",
                                        status_code=status.HTTP_403_FORBIDDEN)
                else:
                    return access
            return authenticate

        if isinstance(self.config.get("KEDRO_GRAPHQL_EVENTS_CONFIG", None), dict):

            @self.post(
                "/event/",
                dependencies=[Depends(authenticate_factory(action="create_event"))],
            )
            async def event(request: Request):
                """
                Endpoint to handle cloudevents.

                Args:
                request (Request): The incoming request containing the CloudEvent.
                Returns:
                dict: Result id of the event handling.
                """
                econf = self.config["KEDRO_GRAPHQL_EVENTS_CONFIG"]

                body = await request.body()
                event: CloudEvent = from_http(
                    request.headers, body
                )  # will raise if not a valid cloudevent

                logger.info(f"Received event: {to_json(event)}")

                # match event details to corresponding pipelines
                source = event.get_attributes().get("source", None)
                type = event.get_attributes().get("type", None)

                pipeline_names = [
                    k
                    for k, v in econf.items()
                    if v["source"] == source and v["type"] == type
                ]

                created_pipelines = []

                for n in pipeline_names:

                    pipeline_input = PipelineInput.from_event(
                        name=n, event=event, state="READY"
                    )

                    q_create = build_graphql_query(
                        "createPipelineReturnFull", fragments=["FullPipeline"]
                    )

                    # Need to STAGE to get a pipeline id to pass as parameter
                    resp = await self.schema.execute(
                        q_create,
                        variable_values={
                            "pipeline": pipeline_input.encode(encoder="graphql")
                        },
                        context_value={"request": request},
                    )

                    staged = Pipeline.decode(
                        resp.data["createPipeline"], decoder="graphql"
                    )

                    pipeline_input.state = "READY"

                    pipeline_input.parameters.append(
                        ParameterInput(name="id", value=str(staged.id), type="STRING")
                    )

                    q_update = build_graphql_query(
                        "updatePipelineReturnFull", fragments=["FullPipeline"]
                    )

                    resp = await self.schema.execute(
                        q_update,
                        variable_values={
                            "id": staged.id,
                            "pipeline": pipeline_input.encode(encoder="graphql"),
                        },
                        context_value={"request": request},
                    )

                    created = Pipeline.decode(
                        resp.data["updatePipeline"], decoder="graphql"
                    )
                    created_pipelines.append(created.encode(encoder="dict"))

                    logger.info(
                        f"event " f"{event.get('id')} triggered pipeline {created.id}")

                return created_pipelines

        else:
            logger.warning(
                "KEDRO_GRAPHQL_EVENTS_CONFIG is not set or not a dictionary. "
                "Event handling endpoint will not be available."
            )

        @self.get("/download", dependencies=[Depends(authenticate_factory(action="read_dataset"))])
        def download(token: str):
            """
            Endpoint to download a file.

            Args:
                token (str): The JWT token for authentication.

            Returns:
                FileResponse: The file to download.
            """
            try:
                payload = jwt.decode(
                    token,
                    self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                    algorithms=[
                        self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]]
                )
                path = Path(payload["filepath"])
            except jwt.ExpiredSignatureError:
                raise HTTPException(status_code=403, detail="Token expired")
            except jwt.InvalidTokenError:
                raise HTTPException(status_code=403, detail="Invalid token")

            ALLOWED_ROOTS = []

            for root in self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS"]:
                ALLOWED_ROOTS.append(Path(root).resolve())
                # print(f"Allowed root: {Path(root).resolve()}")

            if not any(path.is_relative_to(root) for root in ALLOWED_ROOTS):
                raise HTTPException(
                    status_code=403,
                    detail=f"Path {path} is not allowed. Allowed roots: {ALLOWED_ROOTS}"
                )

            if not path.exists() or not path.is_file():
                raise HTTPException(status_code=404, detail="File not found")

            return FileResponse(
                path,
                headers={"Cache-Control": "no-store", "Pragma": "no-cache", "Expires": "0",
                         "Access-Control-Allow-Origin": "*"})

        @self.post("/upload", dependencies=[Depends(authenticate_factory(action="create_dataset"))])
        async def upload(token: str = Form(...), file: UploadFile = File(...)):
            """
            Endpoint to upload a file.

            Args:
                token (str): The JWT token for authentication.
                file (UploadFile): The file to upload.

            Returns:
                dict: A success message and the file path.
            """

            if file.size > self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB"] * 1024 * 1024:
                raise HTTPException(
                    status_code=400,
                    detail=f"File size exceeds the maximum limit of {self.config['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB']} MB"
                )

            try:
                payload = jwt.decode(token, self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                                     algorithms=[self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"]])
                path = Path(payload["filepath"]).resolve()
                # print(f"Destination path: {path}")
            except jwt.ExpiredSignatureError:
                raise HTTPException(status_code=403, detail="Token expired")
            except jwt.InvalidTokenError:
                raise HTTPException(status_code=403, detail="Invalid token")

            ALLOWED_ROOTS = []

            for root in self.config["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS"]:
                ALLOWED_ROOTS.append(Path(root).resolve())
                # print(f"Allowed root: {Path(root).resolve()}")

            if not any(path.is_relative_to(root) for root in ALLOWED_ROOTS):
                raise HTTPException(
                    status_code=403,
                    detail=f"Path {path} is not allowed. Allowed roots: {ALLOWED_ROOTS}"
                )

            try:
                path.parent.mkdir(parents=True, exist_ok=True)
                with open(path, "wb") as out_file:
                    shutil.copyfileobj(file.file, out_file)
            except Exception as e:
                raise HTTPException(status_code=500, detail=f"Upload failed: {e}")

            return {"status": "success", "path": str(path)}



import pytest

import json
from ..utilities import kedro_graphql_config
from kedro_graphql.signed_url.s3_provider import S3Provider
from kedro_graphql.models import DataSet, SignedUrl, SignedUrls


class TestS3Provider:
    config = kedro_graphql_config()

    @pytest.fixture
    def mock_s3_client(self, mocker):
        """Fixture to mock boto3 S3 client"""
        mock_boto3 = mocker.patch("kedro_graphql.signed_url.s3_provider.boto3")
        mock_s3 = mock_boto3.client.return_value
        mock_s3.generate_presigned_url.return_value = "https://your-bucket-name.s3.amazonaws.com/your-object-key?AWSAccessKeyId=your-access-key-id&Signature=your-signature&x-amz-security-token=your-security-token&Expires=expiration-time"
        mock_s3.generate_presigned_post.return_value = {
            "url": "https://your-bucket-name.s3.amazonaws.com/",
            "file": "your-object-key",
            "fields": {
                "key": "your-object-key",
                "AWSAccessKeyId": "YOUR_ACCESS_KEY_ID",
                "policy": "YOUR_BASE64_ENCODED_POLICY",
                "signature": "YOUR_SIGNATURE",
                "Content-Type": "application/octet-stream"
            }
        }
        return mock_s3

    @pytest.fixture
    def mock_dataset(self):
        """Fixture for a TextDataset pointing to a test data file"""

        return DataSet(name="test_dataset", config=json.dumps(
            {"type": "text.TextDataset", "filepath": "s3://my-bucket/path/to/file.txt"}))

    @pytest.fixture
    def mock_partitioned_dataset(self):
        """Fixture for a PartitionedDataset pointing to test partitions"""
        return DataSet(name="test_partitioned_dataset", config=json.dumps(
            {"type": "partitions.PartitionedDataset",
             "path": "s3://my-bucket/path/to/partitioned_dataset",
             "filename_suffix": ".txt",
             "dataset": {"type": "text.TextDataset"}}
        ))

    def test_read(self, mock_s3_client, mock_info_context, mock_dataset):
        """Test reading a single file from a TextDataset"""

        output = S3Provider.read(
            mock_info_context, mock_dataset, expires_in_sec=10)

        assert isinstance(output, SignedUrl)

        assert output.url == "https://your-bucket-name.s3.amazonaws.com/your-object-key?AWSAccessKeyId=your-access-key-id&Signature=your-signature&x-amz-security-token=your-security-token&Expires=expiration-time"

    def test_read_partitions(self, mock_s3_client, mock_info_context, mock_partitioned_dataset):
        """Test reading multiple partitions from a PartitionedDataset"""

        output = S3Provider.read(
            mock_info_context, mock_partitioned_dataset, expires_in_sec=10, partitions=["part-0001", "part-0002"])
        assert isinstance(output, SignedUrls)
        for url in output.urls:
            assert url.url == "https://your-bucket-name.s3.amazonaws.com/your-object-key?AWSAccessKeyId=your-access-key-id&Signature=your-signature&x-amz-security-token=your-security-token&Expires=expiration-time"

    def test_create(self, mock_s3_client, mock_info_context, mock_dataset):
        """Test creating a single signed URL for uploading a TextDataset"""

        output = S3Provider.create(
            mock_info_context, mock_dataset, expires_in_sec=10)

        assert isinstance(output, SignedUrl)

        assert output == SignedUrl.decode({
            "url": "https://your-bucket-name.s3.amazonaws.com/",
            "file": "file.txt",
            "fields": [
                {"name": "key", "value": "your-object-key"},
                {"name": "AWSAccessKeyId", "value": "YOUR_ACCESS_KEY_ID"},
                {"name": "policy", "value": "YOUR_BASE64_ENCODED_POLICY"},
                {"name": "signature", "value": "YOUR_SIGNATURE"},
                {"name": "Content-Type", "value": "application/octet-stream"}
            ]
        }, decoder="graphql")

    def test_create_partitions(self, mock_s3_client, mock_info_context, mock_partitioned_dataset):
        """Test creating signed URLs for the upload of specific partitions in a PartitionedDataset"""

        output = S3Provider.create(
            mock_info_context, mock_partitioned_dataset, expires_in_sec=10, partitions=["part-0001", "part-0002"])
        assert isinstance(output, SignedUrls)

        expected = SignedUrls(urls=[
            SignedUrl.decode({
                "url": "https://your-bucket-name.s3.amazonaws.com/",
                "file": "part-0001.txt",
                "fields": [
                    {"name": "key", "value": "your-object-key"},
                    {"name": "AWSAccessKeyId", "value": "YOUR_ACCESS_KEY_ID"},
                    {"name": "policy", "value": "YOUR_BASE64_ENCODED_POLICY"},
                    {"name": "signature", "value": "YOUR_SIGNATURE"},
                    {"name": "Content-Type", "value": "application/octet-stream"}
                ]
            }, decoder="graphql"),
            SignedUrl.decode({
                "url": "https://your-bucket-name.s3.amazonaws.com/",
                "file": "part-0002.txt",
                "fields": [
                    {"name": "key", "value": "your-object-key"},
                    {"name": "AWSAccessKeyId", "value": "YOUR_ACCESS_KEY_ID"},
                    {"name": "policy", "value": "YOUR_BASE64_ENCODED_POLICY"},
                    {"name": "signature", "value": "YOUR_SIGNATURE"},
                    {"name": "Content-Type", "value": "application/octet-stream"}
                ]
            }, decoder="graphql")
        ])

        for index, item in enumerate(output.urls):
            assert item == expected.urls[index]



import pytest
from kedro_graphql.models import DataSetInput, PipelineInput, Pipeline, TagInput, SignedUrl, SignedUrls
import json
from celery.states import ALL_STATES
from kedro_graphql.schema import encode_cursor
import pytest_asyncio


@pytest_asyncio.fixture
async def mock_create_pipeline(mock_client, mock_text_in, mock_text_out):

    input_dict = {"type": "text.TextDataset", "filepath": str(mock_text_in)}
    output_dict = {"type": "text.TextDataset", "filepath": str(mock_text_out)}

    pipeline_input = PipelineInput(**{
        "name": "example00",
        "state": "READY",
        "data_catalog": [{"name": "text_in", "config": json.dumps(input_dict)},
                         {"name": "text_out", "config": json.dumps(output_dict)}],
        "parameters": [{"name": "example", "value": "hello"},
                       {"name": "duration", "value": "0", "type": "FLOAT"}],
        "tags": [{"key": "author", "value": "opensean"},
                 {"key": "package", "value": "kedro-graphql"}]
    })
    expected = Pipeline.decode(pipeline_input)
    pipeline = await mock_client.create_pipeline(pipeline_input)
    return pipeline_input, expected, pipeline


@pytest_asyncio.fixture
async def mock_create_pipeline_staged(mock_client, mock_text_in, mock_text_out):

    input_dict = {"type": "text.TextDataset", "filepath": str(mock_text_in)}
    output_dict = {"type": "text.TextDataset", "filepath": str(mock_text_out)}

    pipeline_input = PipelineInput(**{
        "name": "example00",
        "state": "STAGED",
        "data_catalog": [{"name": "text_in", "config": json.dumps(input_dict)},
                         {"name": "text_out", "config": json.dumps(output_dict)}],
        "parameters": [{"name": "example", "value": "hello"},
                       {"name": "duration", "value": "0", "type": "FLOAT"}],
        "tags": [{"key": "author", "value": "opensean"},
                 {"key": "package", "value": "kedro-graphql"}]
    })

    expected = Pipeline.decode(pipeline_input)
    pipeline = await mock_client.create_pipeline(pipeline_input)
    return pipeline_input, expected, pipeline


@pytest_asyncio.fixture
async def mock_create_pipeline_staged_with_partitions(mock_client):

    config_str = json.dumps(
        {"type": "partitions.PartitionedDataset",
         "path": "/tmp/my-bucket/path/to/partitioned_dataset",
         "filename_suffix": ".txt",
         "dataset": {"type": "text.TextDataset"}}
    )

    pipeline_input = PipelineInput(**{
        "name": "example00",
        "state": "STAGED",
        "data_catalog": [{"name": "text_in", "config": config_str}],
        "parameters": [{"name": "example", "value": "hello"},
                       {"name": "duration", "value": "0", "type": "FLOAT"}],
        "tags": [{"key": "author", "value": "opensean"},
                 {"key": "package", "value": "kedro-graphql"}]
    })

    expected = Pipeline.decode(pipeline_input)
    pipeline = await mock_client.create_pipeline(pipeline_input)
    return pipeline_input, expected, pipeline


@pytest_asyncio.fixture
async def mock_create_pipeline_staged_unique(mock_client, mock_text_in, mock_text_out):

    input_dict = {"type": "text.TextDataset", "filepath": str(mock_text_in)}
    output_dict = {"type": "text.TextDataset", "filepath": str(mock_text_out)}

    pipeline_input = PipelineInput(**{
        "name": "example00",
        "state": "STAGED",
        "data_catalog": [{"name": "text_in", "config": json.dumps(input_dict)},
                         {"name": "text_out", "config": json.dumps(output_dict)}],
        "parameters": [{"name": "example", "value": "hello"},
                       {"name": "duration", "value": "0", "type": "FLOAT"}],
        "tags": [{"key": "author", "value": "opensean"},
                 {"key": "package", "value": "kedro-graphql"},
                 {"key": "unique", "value": "unique"}]
    })

    expected = Pipeline.decode(pipeline_input)
    pipeline = await mock_client.create_pipeline(pipeline_input)
    return pipeline_input, expected, pipeline


class TestKedroGraphqlClient:

    @pytest.mark.asyncio
    async def test_create_pipeline(self, mock_create_pipeline_staged):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged

        assert pipeline.name == expected.name
        assert pipeline.data_catalog == expected.data_catalog
        assert pipeline.parameters == expected.parameters
        assert pipeline.tags == expected.tags

    @pytest.mark.asyncio
    async def test_read_pipeline(self, mock_create_pipeline_staged, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged
        r = await mock_client.read_pipeline(id=pipeline.id)
        assert r.name == expected.name
        assert r.data_catalog == expected.data_catalog
        assert r.parameters == expected.parameters
        assert r.tags == expected.tags

    @pytest.mark.asyncio
    async def test_read_pipelines(self, mock_create_pipeline_staged_unique, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged_unique
        limit = 1
        filter = "{\"tags.key\": \"unique\", \"tags.value\": \"unique\"}"
        cursor = encode_cursor(pipeline.id)
        sort = "[(\"created_at\", -1)]"
        r = await mock_client.read_pipelines(filter=filter, limit=limit, cursor=cursor, sort=sort)
        assert r.pipelines[0].name == expected.name
        assert r.pipelines[0].data_catalog == expected.data_catalog
        assert r.pipelines[0].parameters == expected.parameters
        assert r.pipelines[0].tags == expected.tags

    @pytest.mark.asyncio
    async def test_update_pipeline(self, mock_create_pipeline_staged, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged
        pipeline_input.tags.append(TagInput(key="test-update", value="updated"))
        r = await mock_client.update_pipeline(id=pipeline.id, pipeline_input=pipeline_input)
        assert r.tags[-1].key == "test-update"
        assert r.tags[-1].value == "updated"

    @pytest.mark.asyncio
    async def test_delete_pipeline(self, mock_create_pipeline_staged, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged
        r = await mock_client.delete_pipeline(id=pipeline.id)
        assert r.id == pipeline.id

    @pytest.mark.asyncio
    async def test_read_datasets(self, mock_create_pipeline_staged, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged
        r = await mock_client.read_datasets(id=pipeline.id, datasets=[DataSetInput(name="text_in"), DataSetInput(name="text_out")], expires_in_sec=3600)
        assert isinstance(r, list)
        assert len(r) == 2
        assert isinstance(r[0], SignedUrl)

    @pytest.mark.asyncio
    async def test_create_datasets(self, mock_create_pipeline_staged, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged
        r = await mock_client.create_datasets(id=pipeline.id, datasets=[DataSetInput(name="text_in"), DataSetInput(name="text_out")], expires_in_sec=3600)
        assert isinstance(r, list)
        assert len(r) == 2
        assert isinstance(r[0], SignedUrl)

    @pytest.mark.asyncio
    async def test_create_datasets_with_partitions(self, mock_create_pipeline_staged_with_partitions, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline_staged_with_partitions
        r = await mock_client.create_datasets(id=pipeline.id, datasets=[DataSetInput(name="text_in", partitions=["partition1", "partition2"])], expires_in_sec=3600)
        assert isinstance(r, list)
        assert len(r) == 1
        assert isinstance(r[0], SignedUrls)
        assert isinstance(r[0].urls[0], SignedUrl)

    @pytest.mark.asyncio
    async def test_pipeline_events(self, mock_celery_session_app, celery_session_worker, mock_create_pipeline, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline

        # Only consume a limited number of events to avoid hanging
        event_count = 0
        max_events = 5  # Only consume first 5 events then break
        
        async for result in mock_client.pipeline_events(id=pipeline.id):
            assert result.status in ALL_STATES
            event_count += 1
            if event_count >= max_events:
                break

    @pytest.mark.asyncio
    async def test_pipeline_logs(self, mock_celery_session_app, celery_session_worker, mock_create_pipeline, mock_client):

        pipeline_input, expected, pipeline = mock_create_pipeline

        # Only consume a limited number of log messages to avoid hanging
        log_count = 0
        max_logs = 5  # Only consume first 5 log messages then break
        
        async for result in mock_client.pipeline_logs(id=pipeline.id):
            assert result.id == pipeline.id
            log_count += 1
            if log_count >= max_logs:
                break



import param
import panel as pn
from kedro_graphql.ui.decorators import UI_PLUGINS


class PipelineCards(pn.viewable.Viewer):
    """
    A component that displays cards for each pipeline with registered @ui_form plugins.
    This component allows users to navigate to a form for running a pipeline or exploring it.

    Attributes:
        spec (dict): The specification for the UI, including configuration and pages.
        explore_page (str): The page to navigate to when a pipeline is explored.
        form_page (str): The page to navigate to when a form is selected.
    """
    spec = param.Dict(default={})
    explore_page = param.String(
        default="explore", doc="The page to navigate to when a pipeline is explored.")
    form_page = param.String(
        default="form", doc="The page to navigate to when a form is selected.")

    def __init__(self, **params):
        super().__init__(**params)

    def navigate(self, button_click, event=None, pipeline=None, form=None):
        """Navigate to the specified page with the given pipeline and form.

        Args:
            button_click (pn.widgets.Button): The button that was clicked.
            event (str): The event that triggered the navigation, either "run" or "explore".
            pipeline (str): The name of the pipeline to navigate to.
            form (panel.viewable.Viewer): The form to navigate to.
        """

        if event == "run":
            pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + self.form_page
            pn.state.location.search = "?pipeline=" + pipeline + "&form=" + form.__name__
            pn.state.location.reload = True
        elif event == "explore":
            pn.state.location.pathname = self.spec["panel_get_server_kwargs"]["prefix"] + self.explore_page
            pn.state.location.search = "?pipeline=" + pipeline
            pn.state.location.reload = True

    @param.depends("explore_page", "form_page")
    async def build_component(self):
        """Builds the pipeline cards component, displaying all pipelines that have a @ui_form plugin registered.

        Yields:
            pn.FlexBox: A flexible box containing cards for each pipeline.
        """
        p_cards = []
        for pipeline, forms in UI_PLUGINS["FORMS"].items():
            p_card = pn.Card(title=pipeline, sizing_mode='stretch_width')
            run_button = pn.widgets.Button(name='Run', button_type='success')
            explore_button = pn.widgets.Button(name='Explore', button_type='default')
            p_card = pn.Card(
                pipeline,
                pn.Row(run_button, explore_button),
                title=pipeline
            )
            p_cards.append(p_card)
            pn.bind(
                self.navigate,
                run_button,
                event="run",
                pipeline=pipeline,
                form=forms[0],
                watch=True
            )
            pn.bind(
                self.navigate,
                explore_button,
                event="explore",
                pipeline=pipeline,
                form=forms[0],
                watch=True
            )

        yield pn.FlexBox(*p_cards)

    def __panel__(self):
        return self.build_component
